#ifndef _IMG_TRANS_MODELS_H
#define _IMG_TRANS_MODELS_H
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
#include <iostream>
#include <iomanip>

#include "imalg_resampling.h"
#include "alexandresku.h"
#include "num_linalg.h"
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
namespace Vtl
{
namespace ImAlg
{
	/*!
	 * \brief    .       Model.
	 *       Model  Resamp        
	 *   .
	 *
	 *****************************************************************************************/
	namespace Transf
	{
		template<typename T>
		class Offset
		{
		public:
			typedef T value_type;
			typedef Vtl::NumAlg::Lin::ublas::c_vector<T,2> VectorB;
			enum { isStrightLineSaving=1};
		private:
			VectorB b;
		public:
			Offset() {}
			Offset(Offset& ot) :b(ot.b) {}
			Offset(const VectorB& b_) :b(b_) {}
			Offset& operator=(const Offset& ot) {b=ot.b;}
			template<typename T1,typename T2>
			void operator()(T1 x,T2 y,T& x_out,T& y_out) const
			{
				x_out=T(x)+b(0);		
				y_out=T(y)+b(1);
			}
			template<typename T1,typename T2>
			VectorB operator()(T1 x,T2 y) const
			{
				VectorB v;
				operator()(x,y,v(0),v(1));
				return v;
			}
			VectorB operator()(VectorB x) const
			{
				return x+b;
			}
		};
		template<typename T>
		class Affine
		{
		public:
			typedef T value_type;
			typedef Vtl::NumAlg::Lin::ublas::c_matrix<T,2,2> MatrixA;
			typedef Vtl::NumAlg::Lin::ublas::c_vector<T,2> VectorB;
			enum { isStrightLineSaving=1};
		private:
			template<typename TT> friend std::ostream& operator<<(std::ostream& os,const Affine<TT>& transf);
		private:
			MatrixA A;
			VectorB b;
		public:
			Affine() {}
			Affine(Affine& at) :A(at.A), b(at.b) {}
			Affine(const MatrixA& A_,const VectorB& b_) :A(A_), b(b_) {}
			Affine& operator=(const Affine& at) {A=at.A; b=at.b;}
			template<typename T1,typename T2>
			void operator()(T1 x,T2 y,T& x_out,T& y_out) const
			{
				VectorB out;
				out=operator()(x,y);
				x_out=out(0);
				y_out=out(1);
			}
			template<typename T1,typename T2>
			VectorB operator()(T1 x,T2 y) const
			{
				VectorB v;
				v(0)=T(x);		v(1)=T(y);
				return operator()(v);
			}
			VectorB operator()(VectorB x) const
			{
				using namespace Vtl::NumAlg::Lin::ublas;
				return prod(A,x)+b;
			}
		};
		template<typename T>
		std::ostream& operator<<(std::ostream& os,const Affine<T>& transf)
		{
			os << "AffineTransf:" << std::endl;
			os << std::setprecision(4) << std::setiosflags(std::ios::showpoint|std::ios::fixed);

			os << "    | "<< std::setw(7) << transf.A(0,0) << "  "<< std::setw(7) << transf.A(0,1) << "|       |" << std::setw(7) << transf.b(0)<< "|" << std::endl;
			os << "  A=| "<< "       "                << "  "<< "       "                << "|     b=|" << "       "    << "|" << std::endl;
			os << "    | "<< std::setw(7) << transf.A(1,0) << "  "<< std::setw(7) << transf.A(1,1) << "|       |" << std::setw(7) << transf.b(1) << "|" << std::endl;
			os << "------------" << std::endl;
			return os;
		}
		template<typename T>
		class Homography
		{
		public:
			typedef T value_type;
			typedef Vtl::NumAlg::Lin::ublas::c_matrix<T,3,3> MatrixH;
			typedef Vtl::NumAlg::Lin::ublas::c_matrix<T,2,2> MatrixA;
			typedef Vtl::NumAlg::Lin::ublas::c_vector<T,2> Vector2;
			typedef Vtl::NumAlg::Lin::ublas::c_vector<T,3> Vector3;
			enum { isStrightLineSaving=1};
		private:
			MatrixH H;
		public:
			Homography() {}
			Homography(Homography& ht) :H(ht.H){}
			Homography(const MatrixH& H) :H(H_) {}
			Homography& operator=(const Homography& ht) {H=ht.H; }
			///offset transform as homography
			Homography(Offset<T>& ot)            { set(ot.b);}
			///offset transform as homography
			Homography(const Vector2& b)         { set(b);}
			///offset transform as homography
			Homography& operator=(Offset<T>& ot) {set(ot.b);}
			///offset transform as homography
			void set(const Vector2& b) 
			{
				using namespace Vtl::NumAlg::Lin::ublas;
				H=identity_matrix<T>(3,3);
				subrange(H, 0,2, 2,3) = b;
			}
			///Affine transform as homography
			Homography(Affine<T>& at) { set(at.A,at.b);}
			///Affine transform as homography
			Homography(const MatrixA& A,const Vector2& b) { set(A,b);}
			///Affine transform as homography
			Homography& operator=(Affine<T>& at) {set(at.A,at.b);}
			///Affine transform as homography
			void set(const MatrixA& A,const Vector2& b) 
			{
				using namespace Vtl::NumAlg::Lin::ublas;
				subrange(H, 0,2, 0,2) = A;
				subrange(H, 0,2, 2,3) = b;
				subrange(H, 2,3, 0,3) = 0;
			}
			template<typename T1,typename T2>
			void operator()(T1 x,T2 y,T& x_out,T& y_out) const
			{
				Vector2 out;
				out=operator()(x,y);
				x_out=out(0);
				y_out=out(1);
			}
			template<typename T1,typename T2>
			Vector2 operator()(T1 x,T2 y) const
			{
				Vector2 v;
				v(0)=T(x);		v(1)=T(y);
				return operator()(v);
			}
			Vector2 operator()(Vector2 x) const
			{
				using namespace Vtl::NumAlg::Lin::ublas;
				Vector3 x3;
				x3(0)=x(0);
				x3(1)=x(1);
				x3(2)=1;
				x3=prod(H,x3);
				x(0)=x3(0)/x3(2);
				x(1)=x3(1)/x3(2);
				return x;
			}
			Vector3 operator()(Vector3 x) const
			{
				using namespace Vtl::NumAlg::Lin::ublas;
				return prod(H,x);
			}
		};
	}//namespace Transf
	/*!
	 * \brief   -      .
	 *          Vtl::ImAlg::Registration::OFlow.
	 * 
	 *      ,     ,    .
	 *  
	 * \code
	 *	template<typename T>
	 *	class Affine
	 *	{
	 *    ...
	 *	};
	 * \endcode
	 *
	 *        , :
	 * \code
	 *  template<typename T>
	 *  class X
	 *  {
	 *  public:
	 *      enum { N=6}; //  ,   ,    .
	 *      enum { isLinear=1};//     (1  0).       
	 *                         //        \ref Lit_KLT_ST,   
	 *                         //     - \ref Lit_Zokai .
	 *      enum { isStrightLineSaving=1}; //      ,    
	 *                                     // ,       , 
	 *                                     //     .
	 *  public:
	 *		//     :
	 *     void setIdentity();
	 *     void Inverse();
	 *     void UpdateByNewAfterThis(Affine& model_new);
	 *     template<class VectorP> void set(VectorP theta_);
	 *     void Interp(T* in,int w, int h,int nls,T* out) const;
	 *     T max_err_at_rect(Affine& model_2,T x1,T y1,T x2,T y2) const;
	 *
	 *  //   
	 *		template<unsigned n> T Dx(T x,T y) const;
	 *		template<unsigned n> T Dy(T x,T y) const;
	 *  //
	 *		template<unsigned n> static T Dx(T x,T y);
	 *		template<unsigned n> static T Dy(T x,T y);
	 *  //     ,  . .
	 * 
	 *	};
	 *
	 *  ostream& operator<<(ostream& os,const X& model); //!<     ,     .
	 *
	 * \endcode
	 *  ,      .      
	 *  ( ),      <b> </b> 
	 *  <b> </b>.
     *
	 *
	 * 
	 * \code
	 *     void setIdentity();
	 * \endcode
	 *    ,     .
	 *
	 * 
	 * \code
	 *     void Inverse();
	 * \endcode
	 *     .
	 *
	 * 
	 * \code
	 *     void UpdateByNewAfterThis(Affine& model_new);
	 * \endcode
	 *     ().     :
	 * \code
	 *   Model A;
	 *   Model B;
	 * \endcode
	 *     \f$T_A , T_B \f$ .   
	 * \f$ \vec r = (x,y)_T\f$           
	 *  \f$ \vec r' = (x',y')_T\f$ .   
	 * \code
	 *   A.UpdateByNewAfterThis(B); 
	 * \endcode
	 *    A    ,  \f$ T_ = T_B T_A \f$ , ..    
	 *  A,   B.
	 *
	 *    ,   () <b>  </b>
	 *
	 * 
	 * \code
	 *     template<class VectorP> void set(VectorP theta_);
	 * \endcode
	 *   Vtl::ImAlg::Registration::OFlow    . VectorP -  
	 *   N  boost::numeric::ublas::c_vector (  (alias) Vtl::NumAlg::Lin::ublas::c_vector ),
	 *  N -   . \n
	 *  OFlow   ,          
	 *  . \n
	 *  ,   theta_ -    ,    . ..    
	 *   theta_  ,     theta_  ,   
	 * . ,   - ,      
	 * \f[ H=\biggl\Vert\begin{array}{ccc}  {A_{11}} & {A_{12}} & {A_{13}}  \\  {A_{21}} & {A_{22}} & {A_{23}} \\  {A_{31}} & {A_{32}} & {A_{33}}  \end{array}\biggr\Vert. \f]\n
	 *    theta_      :
	 * \f[ H=\biggl\Vert\begin{array}{ccc}  {1+ \theta_0} & { \theta_1} & { \theta_2}  \\  { \theta_3} & {1+ \theta_4} & { \theta_5} \\  { \theta_6} & { \theta_7} & {1+ \theta_8}  \end{array}\biggr\Vert. \f]\n
	 *
	 *
	 * 
	 * \code
	 *     void Interp(T* in,int w, int h,int nls,T* out) const;
	 * \endcode
	 *    in,   w,  h     T   nls
	 *    out      (   !),   
	 * <b></b> ,  . ..   ,   
	 *  <b>   </b>.   ,     
	 *      ,       
	 * <b></b>    ,     <b></b> 
	 *          .      
	 * Interp()        ,    .
	 *    ""   ,     ,   
	 *        .
	 *
	 *
	 * 
	 * \code
	 *     T max_err_at_rect(Affine& model_2,T x1,T y1,T x2,T y2) const;
	 * \endcode
	 *      . x1,y1,x2,y2   ( )
	 *     (x,y),       *this  model_2
	 *      ,    .    
	 *        x   y       
	 * .
	 *
	 * 
	 * \code
	 *    template<unsigned n> T Dx(T x,T y) const ;
	 *    template<unsigned n> T Dy(T x,T y) const ;
	 * \endcode
	 *       . ,  
	 * \f$ T:   \vec r = (x,y)_T  ,   \vec r' = (x',y')_T   \vec r' = T(\theta) \vec r \f$ .   
	 *  \f$ \left\lbrace\begin{array}{c}  x' = fx(\theta,x,y)  \\  y' = fy(\theta,x,y) \end{array} \right. \f$ \n
	 * : \n
	 * Dx<n> = \f$ d fx(\theta,x,y) \over d \theta_n \f$ \n
	 * Dy<n> = \f$ d fy(\theta,x,y) \over d \theta_n \f$ \n
	 *
	 *   ,           ,
	 *     , ,       
	 *  .  ,
	 *  ,         -     ,
	 *    - .   ,     ( 0  1), 
	 *    ,   ,    .
	 *   (     n,      
	 * ,      ,    runtime.
	 * 
	 *******************************************************************************************/
	namespace Model
	{
		/*!
		 * \brief  -        .    
		 *        .           
		 *            (  copy-pase).
		 *******************************************************************************************/
		template<typename T>
		class Base
		{
		public:
			typedef T value_type;
			typedef Vtl::NumAlg::Lin::ublas::c_vector<T,2> Vector2;
		protected:
			Base() {}
			~Base() {}
		public:
			T getCof(int w,int h) const
			{
				return T(std::max(w,h)/2.);
			}
			Vector2 getCenter(int w,int h) const
			{
				Vector2 v;
				getCenter(w,h,v);
				return v;
			}
			void getCenter(int w,int h,Vector2& v) const
			{
				v(0)=T(w/2.);
				v(1)=T(h/2.);
			}
		};
		/*!
		 * \brief  -  ,    , (Kanade-Lukas \ref Lit_KLT_TK)
		 *
		 *
		 *******************************************************************************************/
		template<typename T>
		class Offset
			: public Base<T>
		{
		public:
			typedef Vtl::NumAlg::Lin::ublas::c_vector<T,2> VectorB;
			typedef Transf::Offset<T>  Transf;
		public:
			enum { N=2};
			enum { isLinear=1};
			enum { isStrightLineSaving=1};
		public:
			VectorB b;
		public:
			Transf getTransf(int w,int h) const
			{
				return Transf(b*getCof(w,h));
			}
			template<typename Pixel>
			void Interp(const Pixel* in,int w, int h,int nls,Pixel* out,int nls_out,Pixel fill=Pixel(0)) const
			{
				Vtl::ImAlg::Resamp::InterpBilinear<Pixel> interp(in,w,h,nls,fill);
				Vtl::ImAlg::Resamp::Transform(interp,getTransf(w,h),out,w,h,nls_out);
			}
			void setIdentity()
			{
				b[0]=0;
				b[1]=0;
			}
			void Inverse()
			{
				b = -b;
			}
			T max_err_at_rect(Offset& model_2,T x1,T y1,T x2,T y2)
			{
				using namespace Vtl::NumAlg::Lin;
				using namespace Vtl::NumAlg::Lin::ublas;
				VectorB bb=b-model_2.b;
				T err=std::abs( *max_abs_element(bb.begin(),bb.end()) );
				return err;
			}
			template<class VectorP>
			void set(VectorP model_parameters)
			{
				b[0]=T(model_parameters[0]);
				b[1]=T(model_parameters[1]);
			}
			void set(T dx,T dy)
			{
				b[0]=dx;
				b[1]=dy;
			}
			void UpdateByNewAfterThis(Offset& model_new)
			{
				b[0]+=model.b[0];
				b[1]+=model.b[1];
			}
			///         n, 0 < =n < N   x, y
			///  
			/// n=0  ->  1
			/// n=1  ->  1; 
			template<unsigned n>
			T Dx(T x,T y) const { return Vtl::Alg::Return1stArg<(n==0)>(T(1),T(0)); }
			template<unsigned n>
			T Dy(T x,T y) const { return Vtl::Alg::Return1stArg<(n==0)>(T(0),T(1)); }
		};
		template<typename T>
		std::ostream& operator<<(std::ostream& os,const Offset<T>& model)
		{
			os << "ModelOffset:" << std::endl;
			os << std::setprecision(4) << std::setiosflags(std::ios::showpoint|std::ios::fixed);

			os << "       |" << std::setw(7) << model.b[0] << "|" << std::endl;
			os << "     b=|" << "       "                << "|" << std::endl;
			os << "       |" << std::setw(7) << model.b[1] << "|" << std::endl;
			os << "------------" << std::endl;
			return os;
		}
		/*!
		 * \brief  -           
		 *         .    .      
		 *                 (  copy-pase).
		 *******************************************************************************************/
		template<typename T>
		class BaseAffine
			: public Base<T>
		{
		public:
			typedef Vtl::NumAlg::Lin::ublas::c_matrix<T,2,2> MatrixA;
			typedef Vtl::NumAlg::Lin::ublas::c_vector<T,2> VectorB;
			typedef Transf::Affine<T>  Transf;
		public:
			enum { isLinear=1};
			enum { isStrightLineSaving=1};
		public:
			MatrixA A;
			VectorB b;
		protected:
			BaseAffine() {}
			~BaseAffine() {}
			void _setIdentity()
			{
				A(0,0)=1;  A(0,1)=0;    b(0)=0; 
				A(1,0)=0;  A(1,1)=1;    b(1)=0; 
			}
		public:
			Transf getTransf(int w,int h) const
			{
				using namespace Vtl::NumAlg::Lin::ublas;
				T cof=getCof(w,h);
				VectorB v=getCenter(w,h);
				v+=b*cof-prod(A,v);
				return Transf(A,v);
			}
			template<typename Pixel>
			void Interp(const Pixel* in,int w, int h,int nls,Pixel* out,int nls_out,Pixel fill=Pixel(0)) const
			{
				Vtl::ImAlg::Resamp::InterpBilinear<Pixel> interp(in,w,h,nls,fill);
				Vtl::ImAlg::Resamp::Transform(interp,getTransf(w,h),out,w,h,nls_out);
			}
			/*
			 *        
			 *        .
			 ******************************************************************/
			T max_err_at_rect(BaseAffine& model_2,T x1,T y1,T x2,T y2) const
			{
				using namespace Vtl::NumAlg::Lin;
				using namespace Vtl::NumAlg::Lin::ublas;
				VectorB c1; c1(0)=x1; c1(1)=y1; 
				VectorB c2; c2(0)=x1; c2(1)=y2; 
				VectorB c3; c3(0)=x2; c3(1)=y1; 
				VectorB c4; c4(0)=x2; c4(1)=y2; 
				VectorB bb=b-model_2.b;
				MatrixA AA=A-model_2.A;
				c1=prod(AA,c1)+bb;
				c2=prod(AA,c2)+bb;
				c3=prod(AA,c3)+bb;
				c4=prod(AA,c4)+bb;
				T err=std::abs( *max_abs_element(c1.begin(),c1.end()) );
				err=std::max( err , std::abs( *max_abs_element(c2.begin(),c2.end()) ) );
				err=std::max( err , std::abs( *max_abs_element(c3.begin(),c3.end()) ) );
				err=std::max( err , std::abs( *max_abs_element(c4.begin(),c4.end()) ) );
				return err;
			}
		};
		template<typename T>
		std::ostream& operator<<(std::ostream& os,const BaseAffine<T>& model)
		{
			os << "Model BAseAffine:" << std::endl;
			os << std::setprecision(4) << std::setiosflags(std::ios::showpoint|std::ios::fixed);

			os << "    | "<< std::setw(7) << model.A(0,0) << "  "<< std::setw(7) << model.A(0,1) << "|       |" << std::setw(7) << model.b(0)<< "|" << std::endl;
			os << "  A=| "<< "       "                << "  "<< "       "                << "|     b=|" << "       "    << "|" << std::endl;
			os << "    | "<< std::setw(7) << model.A(1,0) << "  "<< std::setw(7) << model.A(1,1) << "|       |" << std::setw(7) << model.b(1) << "|" << std::endl;
			os << "------------" << std::endl;
			return os;
		}
		/*!
		 * \brief  -  ,      y 
		 *        (Shi-Tomasi \ref Lit_ImgRegBushBroom )
		 *
		 *
		 *******************************************************************************************/
		template<typename T>
		class SkewYOffset
			: public BaseAffine<T>
		{
		public:
			enum { N=3};
		public:
			void setIdentity()
			{
				_setIdentity();
			}
			void Inverse()
			{
				using namespace Vtl::NumAlg::Lin;
				using namespace Vtl::NumAlg::Lin::ublas;
				svd::inverse_save(A,A,1.0e-06);
				b = -prod(A,b);
			}
			template<class VectorP>
			void set(VectorP theta)
			{
				set(theta[0],theta[1],theta[2]);
			}
			template<class T1,class T2,class T3>
			void set(T1 skew,T2 dx,T3 dy)
			{
				A(0,0)=T(1);     A(0,1)=T(0);    b(0)=T(dx); 
				A(1,0)=T(skew);  A(1,1)=T(1);    b(1)=T(dy); 
			}
			template<class T2,class T3>
			void offset_by(T2 dx,T3 dy)
			{
				b(0)+=T(dx); 
				b(1)+=T(dy); 
				//Inverse();
			}
			void UpdateByNewAfterThis(SkewYOffset& model_new)
			{
				b=model_new.b+Vtl::NumAlg::Lin::ublas::prod(model_new.A,b);
				A=Vtl::NumAlg::Lin::ublas::prod(model_new.A,A);
			}

			///         n, 0 < =n < N   x, y
			///   
			/// n=0  ->  [ 0 , x ];
			/// n=1  ->  [ 1 , 0 ];
			/// n=2  ->  [ 0 , 1 ]; 
			template<unsigned n>
			T Dx(T x,T y) const { return Vtl::Alg::Return1stArg<(n==1)>(T(1),T(0)); }
			template<unsigned n>
			T Dy(T x,T y) const { return Vtl::Alg::Return1stArg<(n==0)>(x,Vtl::Alg::Return1stArg<(n==2)>(T(1),T(0)) ); }
		};
		/*!
		 * \brief  -  ,      x 
		 *        (Shi-Tomasi \ref Lit_ImgRegBushBroom )
		 *
		 *
		 *******************************************************************************************/
		template<typename T>
		class SkewXOffset
			: public BaseAffine<T>
		{
		public:
			enum { N=3};
		public:
			void setIdentity()
			{
				_setIdentity();
			}
			void Inverse()
			{
				using namespace Vtl::NumAlg::Lin;
				using namespace Vtl::NumAlg::Lin::ublas;
				svd::inverse_save(A,A,1.0e-06);
				b = -prod(A,b);
			}
			template<class VectorP>
			void set(VectorP theta)
			{
				set(theta[0],theta[1],theta[2]);
			}
			template<class T1,class T2,class T3>
			void set(T1 skew,T2 dx,T3 dy)
			{
				A(0,0)=T(1);  A(0,1)=T(skew);    b(0)=T(dx); 
				A(1,0)=T(0);  A(1,1)=T(1);       b(1)=T(dy); 
				//Inverse();
			}
			template<class T2,class T3>
			void offset_by(T2 dx,T3 dy)
			{
				b(0)+=T(dx); 
				b(1)+=T(dy); 
				//Inverse();
			}
			void UpdateByNewAfterThis(SkewXOffset& model_new)
			{
				b=model_new.b+Vtl::NumAlg::Lin::ublas::prod(model_new.A,b);
				A=Vtl::NumAlg::Lin::ublas::prod(model_new.A,A);
			}

			///         n, 0 < =n < N   x, y
			///   
			/// n=0  ->  [ y , 0 ];
			/// n=1  ->  [ 1 , 0 ];
			/// n=2  ->  [ 0 , 1 ]; 
			template<unsigned n>
			T Dx(T x,T y) const { return Vtl::Alg::Return1stArg<(n==0)>(y,Vtl::Alg::Return1stArg<(n==1)>(T(1),T(0)) ); }
			template<unsigned n>
			T Dy(T x,T y) const { return Vtl::Alg::Return1stArg<(n==2)>(T(1),T(0)); }
		};
		/*!
		 * \brief  -  ,      
		 *        (Shi-Tomasi \ref Lit_KLT_ST )
		 *
		 *
		 *******************************************************************************************/
		template<typename T>
		class Similarity
			: public BaseAffine<T>
		{
		public:
			enum { N=4};
		public:
			T scale;
			T phi;
		private:
			void _buildMatrix(T phi_,T scale_)
			{
				A(0,0)=A(1,1)=cos(phi_);
				A(0,1)=sin(phi_);
				A(1,0)=-A(0,1);
				A*=scale_;
			}
		public:
			void setIdentity()
			{
				scale=1;
				phi=0;
				_buildMatrix(phi,scale);
				b(0)=0; 
				b(1)=0; 
			}
			void Inverse()
			{
				scale=1/scale;
				phi=-phi;
				_buildMatrix(phi,scale);
				b = -prod(A,b);
			}
			template<class VectorP>
			void set(VectorP theta_)
			{
				_buildMatrix(theta_[0],theta_[1]);
				b(0)=T(theta_[2]); 
				b(1)=T(theta_[3]); 
				//Inverse();
			}
			void set(T phi_,T scale_,T dx,T dy)
			{
				scale=scale_;
				phi=phi_;
				_buildMatrix(phi_,scale_);
				b(0)=dx; 
				b(1)=dy; 
			}
			void UpdateByNewAfterThis(Similarity& model_new)
			{
				scale*=model_new.scale;
				phi+=model_new.phi;
				_buildMatrix(scale,phi);
				b=model_new.b+Vtl::NumAlg::Lin::ublas::prod(model_new.A,b);
			}
/*
			///         n, 0 < =n < N   x, y
			///   
			/// n=0  ->  [ x , 0 ];
			/// n=1  ->  [ y , 0 ];
			/// n=2  ->  [ 1 , 0 ];
			/// n=3  ->  [ 0 , x ];
			/// n=4  ->  [ 0 , y ];
			/// n=5  ->  [ 0 , 1 ]; 
			template<unsigned n>
			T Dx(T x,T y) const { return Vtl::Alg::Return1stArg<(n<3)>(_D<n>(x,y),T(0)); }
			template<unsigned n>
			T Dy(T x,T y) const { return Vtl::Alg::Return1stArg<(n<3)>(T(0),_D<(n-3)>(x,y)); }
		private:
			template<unsigned n>
			T _D(T x,T y) const { return Vtl::Alg::Return1stArg<(n==0)>(x,Vtl::Alg::Return1stArg<(n==1)>(y,T(1))); }
*/
		};
		/*!
		 * \brief  -  ,      
		 *        (Shi-Tomasi \ref Lit_KLT_ST )
		 *
		 *
		 *******************************************************************************************/
		template<typename T>
		class Affine
			: public BaseAffine<T>
		{
		public:
			enum { N=6};
		public:
			void setIdentity()
			{
				_setIdentity();
			}
			void Inverse()
			{
				using namespace Vtl::NumAlg::Lin;
				using namespace Vtl::NumAlg::Lin::ublas;
				svd::inverse_save(A,A,1.0e-06);
				b = -prod(A,b);
			}
			template<class VectorP>
			void set(VectorP theta_)
			{
				A(0,0)=T(theta_[0]+1);  A(0,1)=T(theta_[1]);    b(0)=T(theta_[2]); 
				A(1,0)=T(theta_[3]);    A(1,1)=T(theta_[4]+1);  b(1)=T(theta_[5]); 
			}
			void UpdateByNewAfterThis(Affine& model_new)
			{
				b=model_new.b+Vtl::NumAlg::Lin::ublas::prod(model_new.A,b);
				A=Vtl::NumAlg::Lin::ublas::prod(model_new.A,A);
			}

			///         n, 0 < =n < N   x, y
			///   
			/// n=0  ->  [ x , 0 ];
			/// n=1  ->  [ y , 0 ];
			/// n=2  ->  [ 1 , 0 ];
			/// n=3  ->  [ 0 , x ];
			/// n=4  ->  [ 0 , y ];
			/// n=5  ->  [ 0 , 1 ]; 
			template<unsigned n>
			T Dx(T x,T y) const { return Vtl::Alg::Return1stArg<(n<3)>(_D<n>(x,y),T(0)); }
			template<unsigned n>
			T Dy(T x,T y) const { return Vtl::Alg::Return1stArg<(n<3)>(T(0),_D<(n-3)>(x,y)); }
		private:
			template<unsigned n>
			T _D(T x,T y) const { return Vtl::Alg::Return1stArg<(n==0)>(x,Vtl::Alg::Return1stArg<(n==1)>(y,T(1))); }
		};
		/*!
		 * \brief  -  ,     
		 *        ( \ref Lit_Zokai )
		 *
		 *
		 *******************************************************************************************/
		template<typename T>
		class Homography
			: public Base<T>
		{
		public:
			typedef T value_type;
			typedef Vtl::NumAlg::Lin::ublas::c_matrix<T,3,3> MatrixA;
			typedef Transf::Homography<T>  Transf;
		public:
			enum { N=8};
			enum { isLinear=0};
			enum { isStrightLineSaving=1};
		public:
			MatrixA A;
		public:
			void Interp(T* in,int w, int h,int nls,T* out) const
			{
			}
			void setIdentity()
			{
				A(0,0)=1;  A(0,1)=0;  A(0,2)=0; 
				A(1,0)=0;  A(1,1)=1;  A(1,2)=0; 
				A(2,0)=0;  A(2,1)=0;  A(2,2)=1; 
			}
			void Inverse() 
			{
				Vtl::NumAlg::Lin::svd::inverse_save(A,A,1.0e-06);
			}
			template<class VectorP>
			void set(VectorP theta_)
			{
				A(0,0)=T(theta_[0]+1);  A(0,1)=T(theta_[1]);    A(0,2)=T(theta_[2]); 
				A(1,0)=T(theta_[3]);    A(1,1)=T(theta_[4]+1);  A(1,2)=T(theta_[5]); 
				A(2,0)=T(theta_[6]);    A(2,1)=T(theta_[7]);    A(2,2)=T(1); 
			}
			void UpdateByNewAfterThis(Homography& model_new)
			{
				A=Vtl::NumAlg::Lin::ublas::prod(model_new.A,A);
			}

			///         n, 0 < =n < N   x, y
			///   . \ref Lit_Zokai
			/// n=0  ->  [ x / (A(2,0)*x + A(2,1)*y + 1) , 0 ]; 
			/// n=1  ->  [ y / (A(2,0)*x + A(2,1)*y + 1) , 0 ];
			/// n=2  ->  [ 1 / (A(2,0)*x + A(2,1)*y + 1) , 0 ];
			/// n=3  ->  [ 0 , x / (A(2,0)*x + A(2,1)*y + 1) ];
			/// n=4  ->  [ 0 , y / (A(2,0)*x + A(2,1)*y + 1) ];
			/// n=5  ->  [ 0 , 1 / (A(2,0)*x + A(2,1)*y + 1) ];
			/// n=6  ->  [ -x*(A(0,0)*x + A(0,1)*y + A(0,2)) / (A(2,0)*x + A(2,1)*y + 1)^2 , -x*(A(1,0)*x + A(1,1)*y + A(1,2)) / (A(2,0)*x + A(2,1)*y + 1)^2 ];
			/// n=7  ->  [ -y*(A(0,0)*x + A(0,1)*y + A(0,2)) / (A(2,0)*x + A(2,1)*y + 1)^2 , -y*(A(1,0)*x + A(1,1)*y + A(1,2)) / (A(2,0)*x + A(2,1)*y + 1)^2 ]; 
			template<unsigned n>
			T Dx(T x,T y) const { return Vtl::Alg::Return1stArg<(n>=6)>(_D3x(x,y), Vtl::Alg::Return1stArg<(n<3)>(_D<n>(x,y),T(0)) ) ; }
			template<unsigned n>
			T Dy(T x,T y) const { return Vtl::Alg::Return1stArg<(n>=6)>(_D3y(x,y), Vtl::Alg::Return1stArg<(n<3)>(T(0),_D<(n-3)>(x,y)) ); }
		private:
			template<unsigned n>
			T _D(T x,T y) const { return Vtl::Alg::Return1stArg<(n==0)>(x/denom(x,y),Vtl::Alg::Return1stArg<(n==1)>(y/denom(x,y),T(1/denom(x,y)))); }
			static T sqr(T v) {return v*v;}
			template<unsigned n>
			T _D3x(T x,T y) const { return Vtl::Alg::Return1stArg<(n==6)>(-x,-y)*T(A(0,0)*x + A(0,1)*y + A(0,2))/sqr(denom(x,y)); }
			template<unsigned n>
			T _D3y(T x,T y) const { return Vtl::Alg::Return1stArg<(n==6)>(-x,-y)*T(A(1,0)*x + A(1,1)*y + A(1,2))/sqr(denom(x,y)); }
			T denom(T x,T y) const { return T(A(2,0)*x + A(2,1)*y + 1);}
		};
		template<typename T>
		std::ostream& operator<<(std::ostream& os,const Homography<T>& model)
		{
			os << "ModelHomography:" << std::endl;
			os << std::setprecision(4) << std::setiosflags(std::ios::showpoint|std::ios::fixed);
			os << "    | "<< std::setw(7) << model.A(0,0) << "  "<< std::setw(7) << model.A(0,1) << "  " << std::setw(7) << model.A(0,2) << "|" << std::endl;
			os << "  A=| "<< std::setw(7) << model.A(1,0) << "  "<< std::setw(7) << model.A(1,1) << "  " << std::setw(7) << model.A(1,2) << "|" << std::endl;
			os << "    | "<< std::setw(7) << model.A(2,0) << "  "<< std::setw(7) << model.A(2,1) << "  " << std::setw(7) << model.A(2,2) << "|" << std::endl;
			os << "------------" << std::endl;
			return os;
		}
	}//namespace Model
}//namespace ImAlg
}//namespace Vtl
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
#endif //_IMG_TRANS_MODELS_H
