#ifndef _STORAGE_H 
#define _STORAGE_H 
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
#include <sstream>
#include <vector>
#include <iostream>
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
namespace Vtl
{
namespace ImAlg
{
namespace Registration
{
/////////////////////////////////////////////////////////////////////////////////////////
	/*! 
	 *  An extract from 
	 *  "Intel  Integrated Performance Primitives (IPP) -Performance Tips and Tricks":
	 *  
	 *  The following performance results using the IPP copy function on an Intel Pentium 4 
	 *  processor clearly present the differences for different alignments of the source data:
	 *  <TABLE>
	 *  <TR><TD> Source align, B </TD><TD> 16+0 </TD><TD>  16+4 </TD><TD> 16+3 </TD></TR>
	 *  <TR><TD> cpe             </TD><TD>  0.3 </TD><TD>  0.5  </TD><TD>  0.7 </TD></TR>
	 *	</TABLE> 
	 *
	 *	The cpe is number of CPU-cycles per element.
	 **********************************************************************************************/
#define PROCESSOR_MEMORY_ALIGNMENT 32
	/*!
	 *  An extract from 
	 *  "Intel  Integrated Performance Primitives (IPP) -Performance Tips and Tricks":
	 *  
	 *  It is very important when developing on the Intel Pentium 4 processor that two pointers used
	 *  by a function do not differ by a multiple of 64 KB. Performance of the function is decreased
	 *  significantly where two pointers differ by a multiple of 64 KB. This is particular to the
	 *  current version of the Intel Pentium 4 processor. For example, consider the
	 *  
	 *  ippsCplxToReal_32fc function:
	 *  \code
	 *  for(n=0;n<length;n++) 
	 *  {
	 *  	pRe[n]=pSrc[n].re;
	 *  	pIm[n]=pSrc[n].im;
	 *  }
	 *  \endcode
	 *  Performance decreases sharply if difference between the pRe and pIm pointers is 64 KB or a
	 *  multiple of that. We can see below that performance becomes 80 cpe instead of 5 cpe.
	 *  \image html intel_64k.jpg 
	 **********************************************************************************************/
#define PROCESSOR_MEMORY_AVOID_PTR_DIFF 64*1024
//#define PROCESSOR_MEMORY_AVOID_PTR_DIFF 0
/////////////////////////////////////////////////////////////////////////////////////////	
	/*!
	 *	\brief ,      (  ).   
	 *         ,  ,          .
	 *          .    PROCESSOR_MEMORY_AVOID_PTR_DIFF .
	 *
	 */
	template<typename T,size_t ndim,size_t align=PROCESSOR_MEMORY_ALIGNMENT,size_t avoid_ptrdiff=PROCESSOR_MEMORY_AVOID_PTR_DIFF>
	class Storage
	{
	public:
		enum {N=ndim};
		typedef size_t Offsets_t[ndim];
	private:
		//size_t ofs[ndim];
		Offsets_t ofs;
		std::vector<T> data_;
		T* data;
	public:
		Storage() :data(0){}
		~Storage() {}
		void Init(Offsets_t sz) // ndim   
		{
			ofs[0]=0;
			for(size_t i=1; i<ndim; ++i)
			{
				size_t d=sz[i-1]*sizeof(T);
				if(align>0)
				{
					size_t e=d % align;
					if(e)
					{
						e=align-e;
						d+=e;
						//std::cout << "Block "<<i<< " aligned by "<<e<<" bytes."<<std::endl;
					}
				}
				ofs[i]=ofs[i-1]+d/sizeof(T);
				if(avoid_ptrdiff>0)
				{
					size_t cnt=ndim;
					bool need;
					do
					{
						need=false;
						for(size_t j=0; j<i; ++j)
						{
							size_t d=(ofs[i]-ofs[j])*sizeof(T);
							if(0==(d%avoid_ptrdiff))
							{
								if(cnt--)
									need=true;
								ofs[i]+=align/sizeof(T);
								//std::cout << "Block "<<i<< " shifted to avoid 64k slowing."<<std::endl;
								break;
							}
						}
					}while(need);
				}
			}
			size_t s=ofs[ndim-1]+sz[ndim-1]+align/sizeof(T);
			data_.resize(s);
			data=&(data_[0]);
			::memset(data,0,sizeof(data[0])*s);

			size_t d=size_t(data);
			if(align>0)
			{
				size_t e=d % align;
				if(e)
					data+=(align-e)/sizeof(T);
			}
		}
		template<size_t i> T* block() { return data+ofs[i];}
		template<size_t i> const T* block() const { return data+ofs[i];}
	};
}//namespace Registration
}//namespace ImAlg
}//namespace Vtl
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
#endif //_STORAGE_H 
