#ifndef DFPMINIMIZATION_H_ 
#define DFPMINIMIZATION_H_ 
#include <math.h>
#include <vector>

// library includings
#include "num_linalg.h"
//#include "nrutil.h"

// own includings
#include "MistakeFunc.h"

namespace Basics3D
{
	#define ITMAX 200				// Maximum allowed number of iterations.
	#define EPS 3.0e-8			// Machine precision.
	#define TOLX (4*EPS)		// Convergence criterion on x values.
	#define STPMX 100.0			// Scaled maximum step length allowed in line searches.
	#define ALF 1.0e-4      //Ensures suffcient decrease in function value.
	//#define TOLX 1.0e-7 Convergence criterion on x.

	/////////////////////////////////////////////////////////////////////////////////////////
	// Given an n-dimensional point xold[1..n], the value of the function and gradient there,
	// fold and g[1..n], and a direction p[1..n], finds a new point x[1..n] along the 
	// direction p from xold where the function func has decreased "suffciently." 
	// The new function value is returned in f. stpmax is an input quantity that limits the
	// length of the steps so that you do not try to evaluate the function in regions where
	// it is undefined or subject to overflow. p is usually the Newton direction. 
	// The output quantity check is false (0) on a normal exit. It is true (1) when x is too
	// close to xold. In a minimization algorithm, this usually signals convergence and can
	// be ignored. However, in a zero-finding algorithm the calling program should check
	// whether the 	convergence is spurious. Some "diffcult" problems may require double
	// precision in this routine.
	template <class type, class MatrixType>
	void lnsrch(int n,
							Vtl::NumAlg::Lin::ublas::vector<type>* xold, //float xold[],
							type fold,
							Vtl::NumAlg::Lin::ublas::vector<type>* g,    //float g[],
							Vtl::NumAlg::Lin::ublas::vector<type>* p,    //float p[],
							Vtl::NumAlg::Lin::ublas::vector<type>* x,    //float x[],
							type *f,
							type stpmax,
							int *check,
							T_MistakeFunc<MatrixType>* Mist)
							
	{
		using namespace Vtl::NumAlg::Lin;

		*check = 0;

		type sum = ublas::inner_prod((*p), (*p));/*
		for (int i = 0; i < n; i++)      //(sum = 0.0, i = 1;i<=n;i++) 
		{
			sum += (*p)[i] * (*p)[i];
		}*/

		sum = sqrt(sum);

		// Scale if attempted step is too big.
		if (sum > stpmax)
		{
			for (int i = 0; i < n; i++)
			{
				(*p)[i] *= stpmax/sum;
			}
		}

		type slope = ublas::inner_prod((*g), (*p));/*0.0;
		for (int i = 0; i < n; i++)  //(slope=0.0,i=1;i<=n;i++)
		{
				slope += (*g)[i] * (*p)[i];
		}*/

		if (slope >= 0.0)
		{
			throw "\nRoundoff problem in lnsrch.";
		}

		//  Compute min.
		type test = 0.0;

		for (int i = 0; i < n; i++)
		{
			type temp = fabs((*p)[i]) / std::max(fabs((*xold)[i]), 1.0);
			
			if (temp > test)
			{
				test = temp;
			}
		}
		
		type alamin = TOLX / test;

		// Always try full Newton step first.
		type alam = 1.0;
		type f2, alam2;

		// Start of iteration loop.
		for (;;) 
		{
			(*x) = (*xold) + alam * (*p);/*
			for (int i = 0; i < n; i++)               //(i=1;i<=n;i++)
			{
				(*x)[i] = (*xold)[i] + alam * (*p)[i];
			}*/

			*f = Mist->Func(*x);

			// Convergence on x. For zero finding,
			// the calling program should verify the convergence.
			if (alam < alamin)
			{
				(*x) = (*xold);/*
				for (int i = 0; i < n; i++)             //(i=1;i<=n;i++)
				{
					(*x)[i] = (*xold)[i];
				}*/
				*check = 1;

				/**/std::cout << "\nCheck = 1, x = xold\n";
				return;
			} 
			else
			{
				type tmplam;
							
				if (*f <= fold + ALF * alam * slope)
				{
					/**/std::cout << "\nSufficient function decrease\n";
					return;  
				}
				else // Backtrack.
				{ 
					if (alam == 1.0) // First time.
					{
						tmplam = -slope / (2.0 * (*f - fold - slope)); 
					}
					else  //Subsequent backtracks.
					{ 
						type rhs1 = *f - fold - alam * slope;
						type rhs2 = f2 - fold - alam2 * slope;
						type a = (rhs1 / (alam * alam) - rhs2 / (alam2 * alam2)) / (alam - alam2);
						type b = (-alam2 * rhs1 / (alam * alam) + alam * rhs2 / (alam2 * alam2)) 
									 / (alam - alam2);
			    
						if (a == 0.0)
						{
							tmplam = -slope / (2.0 * b);
						}
						else 
						{
							type disc = b * b - 3.0 * a * slope;

							if (disc < 0.0)
							{
								tmplam = 0.5 * alam;
							}
							else
							{
								if (b <= 0.0)
								{
									tmplam = (-b + sqrt(disc)) / (3.0 * a);
								}
								else
								{
									tmplam = -slope / (b + sqrt(disc));
								}
							}
						}

						if (tmplam > 0.5 * alam)
						{
							tmplam = 0.5 * alam; //  0:51.
						}
					}
				}

				alam2 = alam;
				f2 = *f;
				alam = std::max(tmplam, 0.1 * alam); //   0:11.
			} // Try again.
		}
	};


	/////////////////////////////////////////////////////////////////////////////////////////
	// Given a starting point p[1..n] that is a vector of length n, 
	// the Broyden-Fletcher-Goldfarb-Shanno variant of Davidon-Fletcher-Powell minimization
	// is performed on a function func, using its gradient as calculated by a routine dfunc.
	// The convergence requirement on zeroing the gradient is input as gtol. 
	// Returned quantities are p[1..n] (the location of the minimum),
	// iter (the number of iterations that were performed), and 
	// fret (the minimum value of the function). 
	// The routine lnsrch is called to perform approximate line minimizations.

	template<class type, class MatrixType>
	void _dfpmin(Vtl::NumAlg::Lin::ublas::vector<type>* p, 
							int n, 
							type gtol, 
							int *iter, 
							type *fret,
							T_MistakeFunc<MatrixType>* Mist)
	{
		using namespace Vtl::NumAlg::Lin;

		ublas::matrix<type> hessin(n,n);
		ublas::vector<type> dg(n);
		ublas::vector<type> g(n);
		ublas::vector<type> hdg(n);
		ublas::vector<type> pnew(n);
		ublas::vector<type> xi(n);
		
		// Calculate starting function value and gradient
		// and initialize the inverse Hessian to the unit matrix.
		type fp = Mist->Func(*p);
		Mist->Derivative(g, *p);

		type sum = 0.0;

		for (int i = 0; i < n; i++)
		{ 
			for (int j = 0; j < n; j++)
			{
				hessin(i,j) = 0.0;
			}
					
			hessin(i,i) = 1.0;
		}

		// Initial line direction.
		xi = -g;
		sum = ublas::inner_prod((*p), (*p));

		type stpmax = STPMX * std::max(sqrt(sum), (type)n);

		// Main loop over the iterations.
		for (int its = 0; its < ITMAX; its++)
		{ 
			*iter = its;

			/**/std::cout << "\niter = " << *iter << "  |  ";
			/**/for (int k = 0; k < n; k++)
			/**/{
			/**/	std::cout << (*p)[k] << "  ";
			/**/}

			int check;
			lnsrch<double>(n, p, fp, &g, &xi, &pnew, fret, stpmax, &check, Mist);

			/**/std::cout << "  |  ";
			/**/for (int k = 0; k < n; k++)
			/**/{
			/**/	std::cout << pnew[k] << "  ";
			/**/}

			// 	The new function evaluation occurs in lnsrch;
			// save the function value in fp for the next line search. 
			// It is usually safe to ignore the value of check.
			fp = *fret;
		
			
			// Update the line direction,
			xi = pnew - (*p);
			// and the current point.
			(*p) = pnew;/*
			for (int i = 0; i < n; i++)               //(i=1;i<=n;i++) 
			{
				xi[i] = pnew[i] - (*p)[i]; 
				(*p)[i] = pnew[i]; 
			}*/
			
			// Test for convergence on x.
			type test = 0.0;

			for (int i = 0; i < n; i++)
			{
				type temp = fabs(xi[i]) / std::max(fabs((*p)[i]), 1.0);
				
				if (temp > test) 
				{
					test = temp;
				}
			}

			if (test < TOLX)
			{
				return;
			}

			// Save the old gradient,
			dg = g;/*
			for (int i = 0; i < n; i++)               //(i=1;i<=n;i++) 
			{
				dg[i] = g[i]; 
			}*/

			// and get the new gradient.
			Mist->Derivative(g, *p);

			// Test for convergence on zero gradient.
			test = 0.0;

			type den = std::max(*fret, 1.0);

			for (int i = 0; i < n; i++)
			{
				type temp = fabs(g[i]) * std::max(fabs((*p)[i]), 1.0) / den;

				if (temp > test) 
				{
					test = temp;
				}
			}

			if (test < gtol)
			{
				return;
			}

			// Compute difference of gradients,
			dg = g - dg;/*
			for (int i = 0; i < n; i++)               //(i = 1; i < n; i++)
			{
				dg[i] = g[i] - dg[i]; 
			}*/

			// and difference times current matrix.
			hdg = ublas::prod(hessin, dg);/*
			for (int i = 0; i < n; i++)               //(i=1;i<=n;i++) 
			{ 
				hdg[i] = 0.0;

				for (int j = 0; j < n; j++)             //(j=1;j<=n;j++) 
				{
					hdg[i] += hessin(i, j) * dg[j];
				}
			}*/
					
			// Calculate dot products for the denominators.
			type fac, fae, sumdg, sumxi;
			fac = fae = sumdg = sumxi = 0.0; 

			fac = ublas::inner_prod(dg, xi);
			fae = ublas::inner_prod(dg, hdg);
			sumdg = ublas::inner_prod(dg, dg);
			sumxi = ublas::inner_prod(xi, xi);/*
			for	(int i = 0; i < n; i++)		            //(i=1;i<=n;i++) 
			{ 
					fac += dg[i] * xi[i];
					fae += dg[i] * hdg[i];
					sumdg += dg[i] * dg[i];
					sumxi += xi[i] * xi[i];
			}*/
					
			// Skip update if fac not suffciently positive.
			if (fac > sqrt(EPS * sumdg * sumxi)) 
			{  
				fac = 1.0 / fac; 
				type fad = 1.0 / fae;

				// The vector that makes BFGS different from DFP:
				dg = fac * xi - fad * hdg;/*
				for (int i = 0; i < n; i++)             //(i=1;i<=n;i++) 
				{
					dg[i] = fac * xi[i] - fad * hdg[i];
				}*/
					
				// The BFGS updating formula:
				hessin += fac * ublas::outer_prod(xi, xi) - fad * ublas::outer_prod(hdg, hdg)
						 + fae * ublas::outer_prod(dg, dg);/*
				for (int i = 0; i < n; i++)             //(i=1;i<=n;i++) 
				{ 
					for (int j = i; j < n; j++)       //(j=i;j<=n;j++) 
					{
						hessin(i, j) += fac * xi[i] * xi[j] - fad * hdg[i] * hdg[j] + fae * dg[i] * dg[j];
						hessin(j, i) = hessin(i, j);
					}
				}*/
			}
					
			// Now calculate the next direction to go,
			xi = - ublas::prod(hessin, g);/*
			for (int i = 0; i < n; i++)
			{ 
				xi[i] = 0.0;

				for (int j = 0; j < n; j++)
				{
					xi[i] -= hessin(i, j) * g[j];
				}
			}*/
			
			// and go back for another iteration.
		} 
				
	//	nrerror("too many iterations in dfpmin");
		
	}
}
#endif //DFPMINIMIZATION_H_ 
