#ifndef _INTEGRAL_TRANSFORMS_H
#define _INTEGRAL_TRANSFORMS_H
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*
 *  : ,  ...
 *
 ******************************************************************************/
#include "numalg.h"
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
namespace Vtl
{
namespace NumAlg
{
//! \brief     
namespace IntTransf
{
	/*! \brief ,       . 
	 *
	 *    ,   (  )    ,
	 *    ,    ,   ,     ..
	 *       -  . ,        \n
	 *   1 2 3   \n
	 *   4 5 6   \n
	 *   7 8 9   \n
	 *   : \n
	 * 1 2 3 4 5 6 7 8 9\n
	 *    -      ,     
	 *       ,   ,   .
	 *    .\n
	 *   ,    -   ,       , 
	 *     (.. 2  ),      -  
	 *   ,    .
     *
	 *
	 *******************************************************************************************/
	namespace Utils
	{
		template<typename T > T Sqr(T v) { return v*v;}

		namespace Impl
		{
			template<typename I,size_t ndim>
			struct ArrayElemProduct
			{
				static I value(I n[])
				{
					return n[ndim-1]*ArrayElemProduct<I,ndim-1>::value(n);
				}
			};
			template<typename I>
			struct ArrayElemProduct<I,0>
			{
				static I value(I n[])
				{
					return 1;
				}
			};
			/*!
			 * Implements trigonometric recurrence.
			 *  \f$ W_k=W^k\f$ ,\f$ W=e^{2\pi i /N} \f$ ,  \f$ k=0,1..N-1 \f$. 
			 *  \c wr  \c wi       W_k. 
			 *     - W=1 ( k=0).
			 *     next()  \f$ W_{k+1}=W_k*W \f$ ,    
			 *    .
			 *  MultR(R re,R im)  MultI(R re,R im)    
			 * C=(re,im)    \f$ W_k=(wr,wi) \f$ ,     
			 *    : \f$ MultR() = Re( C \dot W_k ) \f$ , \f$ MultI() = Im( C \dot W_k ) \f$.
			 *
			 *      .
			 **************************************************************************************/
			template<typename R,typename T,typename I>
			class TrigRecurrence
			{
				const T theta;
				const T wpr;
				const T wpi;
				T wr;
				T wi;
			public:
				//! Initialize the trigonometric recurrence.
				TrigRecurrence(bool reverse,I mmax) 
					: theta((reverse ?  -2*M_PI : 2*M_PI)/mmax)
					 ,wpr(-2.0*Sqr(sin(0.5*theta)))
					 ,wpi(sin(theta))
				{
					wr=1.0;
					wi=0.0;
				}
				//T c() const {return wr;} 
				//T s() const {return wi;} 
				~TrigRecurrence() {} 
				R MultR(R re,R im) const { return R(wr*re-wi*im); }
				R MultI(R re,R im) const { return R(wr*im+wi*re); }
				//! Reccurence step
				void next()
				{
					T wtemp=wr;
					wr+=wtemp*wpr-wi*wpi; //Trigonometric recurrence.
					wi+=wi*wpr+wtemp*wpi;
				}
				void butterfly_fft(R* re,R* im,I i,I j)
				{
					R tempr=MultR(re[j],im[j]);
					R tempi=MultI(re[j],im[j]);
					re[j] = re[i] - tempr;
					im[j] = im[i] - tempi;
					re[i] += tempr;
					im[i] += tempi;
				}
				void butterfly_Hartley(R* even,R* odd,I i,I j)
				{
					//wr=c=cos = (R)cos(2*M_PI*i/blocksize);
					//wi=s=sin = (R)sqrt(1-c*c);
					R d0 = R(wr*odd[i] + wi*odd[j]);
					R d1 = R(wr*odd[j] - wi*odd[i]);
					odd[i] = even[i] - d0;
					odd[j] = even[j] + d1;
					even[i] += d0;
					even[j] += - d1;
				}
			};
		};
		//!     \c n  \c ndim.
		template<size_t ndim,typename I>  
		I ArrayElemProduct(I n[])
		{
			return Impl::ArrayElemProduct<I,ndim>::value(n);
		}
		//!   k ,  2^k >= size 
		template<typename I> 
		I smallest_log2_ge_size(I size)
		{
			I ls=0;//ls = log(size)
			while( (I(1)<<ls) < size ) ++ls;
			return ls;
		}
		//!   ,   2    size 
		template<typename I> 
		I CalcSmalestPow2(I size)
		{
			return 1<<smallest_log2_ge_size(size);
		}

		//!    \ index   ,   \ log2_size - 
		template<typename I> 
		I rev(I index, I log2_size)
		{
			I j = 0;
			while (log2_size--)
			{
				j = (j<<1) | (index & 1);
				index >>= 1;
			}
			return j;
		}
		//!    \ index   ,   \ log2_size - ,
		//! log2(step)       ,     log2_size 
		//!  ,  log2(step).    
		//! \code
		//! I lstep=log2(step);
		//! index >>= lstep;
		//! index=bit_reversal(index,log2_size);
		//! index <<= lstep;
		//! \endcode
		//!   .\n
		//! \b <tt> step    ,   !</tt>
		template<typename I> 
		I rev(I index, I log2_size,I step)
		{
			I j = 0;
			//I step=1<<log2_step;
			while (log2_size--)
			{
				j = (j<<1) | (index & step);
				index >>= 1;
			}
			return j;
		}
		//! \brief      \c data. size    2.
		//!
		//!    i       j,
		//!    i    j,    . \n
		//! \b :    (n)  8.    3 (log(n)) .
		//! : <tt> \n
		//!  i = 0 = 000 <=> j = 0 = 000 \n
		//!  i = 1 = 001 <=> j = 4 = 100 \n
		//!  i = 2 = 010 <=> j = 2 = 010 \n
		//!  i = 3 = 011 <=> j = 6 = 110 \n
		//!  i = 4 = 100 <=> j = 1 = 001 \n
		//!  i = 5 = 101 <=> j = 5 = 101 \n
		//!  i = 6 = 110 <=> j = 3 = 011 \n
		//!  i = 7 = 111 <=> j = 7 = 111 \n
		//!  </tt>
		//! T -   ( float*) ,   operator[](I i),  I    0 <= i < size ; 
		//!    std::swap()      data[i], data[j] . 
		template<typename T,typename I>
		void bit_reversal(T data, I size)
		{
			if(size <= 2) return;
			I ls=smallest_log2_ge_size(size);
			assert( size == (1<<ls) );
			for(I i = 1; (--size) > 0 ; ++i)
			{
				I j = rev(i,ls);
				if (j > i)
					std::swap(data[i], data[j]);
			}
		}
		template<typename T,typename I>
		void bit_reversal(T data, I size, I step)
		{
			if(size <= 2) return;
			I ls=smallest_log2_ge_size(size);
			assert( size == (1<<ls) );
			assert( step == (1<<smallest_log2_ge_size(step)) );
			for(I i = step; (--size) > 0; i+=step)
			{
				I j = rev(i,ls,step);
				if (j > i)
				{
					std::swap(data[i], data[j]);
				}
			}
		}
		template<typename T,typename I>
		void bit_reversal_complex(T data, I size,I step)
		{
			bit_reversal_complex(data, data+1,size,step);
		}
		template<typename T,typename I>
		void bit_reversal_complex(T re, T im, I size,I step)
		{
			if(size <= 2) return;
			I ls=smallest_log2_ge_size(size);
			assert( size == (1<<ls) );
			assert( step == (1<<smallest_log2_ge_size(step)) );
			for(I i = step; (--size) > 0; i+=step)
			{
				I j = rev(i,ls,step);
				if (j > i)
				{
					std::swap(re[i], re[j]);
					std::swap(im[i], im[j]);
				}
			}
		}
		namespace Impl
		{
			template<typename T,typename I>	
			void SwapArrayHalfs(T* data, I size,I step,Loki::Int2Type<false>)
			{
				I ofs=size>>1;
				size=ofs;
				ofs*=step;
				for (I i = 0 ; (size--)!=0; i+=step)
					std::swap(data[i], data[i+ofs]);
			}
			template<typename T,typename I>	
			void SwapArrayHalfs(T* data, I size,I step,Loki::Int2Type<true>)
			{
				I ofs=size>>1;
				size=ofs;
				ofs*=step;
				for (I i = 0 ; (size--)!=0; i+=step)
				{
					std::swap(data[i], data[i+ofs]);//Re
					std::swap(data[i+1], data[i+1+ofs]);//Im
				}
			}
		}
		//!    \c re  \c im       .
		template<typename T,typename I>	
		void SwapArrayHalfs(T* re,T* im, I size,I step)
		{
			I ofs=size>>1;
			size=ofs;
			ofs*=step;
			for (I i = 0 ; (size--)!=0; i+=step)
			{
				std::swap(re[i], re[i+ofs]);
				std::swap(im[i], im[i+ofs]);
			}
		}
		//!    data       
		template<bool is_complex,typename T,typename I>	
		void SwapArrayHalfs(T* data, I size,I step)
		{
			Impl::SwapArrayHalfs(data, size,step,Loki::Int2Type<is_complex>());
		}
		template<typename T,typename I>
		void swap(T* d1,T* d2,I len)
		{
			for(I j = 0; j < len; j++)
				std::swap(d1[j],d2[j]);
		}
		//!   \c n   \c data   \c step
		template<typename T,typename I>
		void mult(T data[],I n,T multiplier,I step=1)
		{
			for( ; n-- ; data+=step )
			{
				(*data)*=multiplier;
			}
		}
		/*!
		 *   ()      ,     .
		 * w  h  (cols)   (rows) ,       ,
		 *    / -  2.
		 */
		template<bool is_complex_pairs,typename T,typename I>
		void SwapQuarters(T data[],I w,I h)
		{
			assert( w==CalcSmalestPow2(w) );
			assert( h==CalcSmalestPow2(h) );
			I n[]={w,h};
			SwapQuadrants<2,is_complex_pairs>(data,n);
			/*
			I w2=w>>1;
			I h2=h>>1;
			I half=w*h2;
			while(;h2--;data+=w)
			{
				T* d2=data+half;
				//     
				swap(data,d2+w2,w2);
				//     
				swap(data+w2,d2,w2);
			}
			*/
		}
		namespace Impl
		{
			template<bool is_complex_pairs,int nd>
			struct SwapQuadrants
			{
				template<typename T,typename I>
				static void sw(T* data_1,T* data_2,I n[],I ntot)
				{
					I nc=n[nd-1];
					ntot/=nc;
					for( nc>>=1 ; nc-- ; data_1+=ntot, data_2+=ntot )
						SwapQuadrants<is_complex_pairs,nd-1>::sw(data_1,data_2,n,ntot);
				}
			};
			template<bool is_complex_pairs>
			struct SwapQuadrants<is_complex_pairs,1>
			{
				template<typename T,typename I>
				static void sw(T* data_1,T* data_2,I n[],I ntot)
				{
					I nc2=is_complex_pairs ? n[0] : n[0]>>1;
					for(I j=0; j< nc2 ; ++j)
						std::swap(data_1[j],data_2[j]);
				}
			};
			//! Used ONLY by IntTransf::Utils::SwapQuadrants() function
			template<unsigned ndim,bool is_complex_pairs,typename I>
			size_t SwapQuadrants_offset(size_t mask,I nn[])
			{
				size_t ofs=0;
				for(size_t  k=0, j=1; k<ndim; ++k, mask>>=1)
				{
					j*=nn[k];
					ofs+=(mask&1)? j/2 : 0;
				}
				return is_complex_pairs? ofs<<1 : ofs;
			}
		}
		/*!
		 * \brief   ,     /   
		 *  .     . \c data -  ,    
		 * ,     ..   \c ndim    k, 
		 * 0 <= k < ndim    n[k].      n[k]    2,
		 *      ,  n[k]     k=0..ndim-1.
		 *
		 * :\n
		 *    ,       k=0..ndim-1,   
		 *    n[k]/2 - .        
		 * ndim .      k ,     0     k , 
		 *   1  ( ),        ,   , 
		 *    xxxxb ,  ndim     ,    
		 *  .   ,        
		 *   0  11..1b (   ndim).      
		 *    0  1<<(ndim-1)          .
		 *     data     mask    
		 * \code
		 * size_t offset=0;
		 * for(size_t  k=0, j=1; k<ndim; ++k, mask>>1)
		 * {
		 *		offset+=(mask&1)? () (n[k]/2)*j : 0;
		 *		j*=n[k];
		 * }
		 * \endcode
		 * \see SwapQuadrants(T re[],T im[],I n[])
		 ************************************************************************************************/
		template<unsigned ndim,bool is_complex_pairs,typename T,typename I>
		void SwapQuadrants(T data[],I n[])
		{
			
			size_t max_mask=(1<<ndim)-1;
			I ntot=(is_complex_pairs ? 2 :1 )*ArrayElemProduct<ndim>(n);
			for(size_t  mask=0 ; mask <= (max_mask>>1); ++mask )
			{
				size_t mask2=mask^max_mask; //(~mask)&max_mask;
				//swap quadrants mask and mask2
				size_t ofs_1=Impl::SwapQuadrants_offset<ndim,is_complex_pairs>(mask,n);
				size_t ofs_2=Impl::SwapQuadrants_offset<ndim,is_complex_pairs>(mask2,n);
				Impl::SwapQuadrants<is_complex_pairs,ndim>::sw(data+ofs_1,data+ofs_2,n,ntot);
			}
		}
		/*!
		 * \brief   ,     /   
		 *  .     . \c data -  ,    
		 * ,     ..   \c ndim    k, 
		 * 0 <= k < ndim    n[k].      n[k]    2,
		 *      ,  n[k]     k=0..ndim-1.
		 * \see SwapQuadrants(T data[],I n[])
		 ************************************************************************************************/
		template<unsigned ndim,typename T,typename I>
		void SwapQuadrants(T re[],T im[],I n[])
		{
			
			size_t max_mask=(1<<ndim)-1;
			const int is_complex_pairs =false;
			I ntot=(is_complex_pairs ? 2 :1 )*ArrayElemProduct<ndim>(n);
			for(size_t  mask=0 ; mask <= (max_mask>>1); ++mask )
			{
				size_t mask2=mask^max_mask; //(~mask)&max_mask;
				//swap quadrants mask and mask2
				size_t ofs_1=Impl::SwapQuadrants_offset<ndim,is_complex_pairs>(mask,n);
				size_t ofs_2=Impl::SwapQuadrants_offset<ndim,is_complex_pairs>(mask2,n);
				Impl::SwapQuadrants<is_complex_pairs,ndim>::sw(re+ofs_1,re+ofs_2,n,ntot);
				Impl::SwapQuadrants<is_complex_pairs,ndim>::sw(im+ofs_1,im+ofs_2,n,ntot);
			}
		}
		//! \brief   ,        data  .
		template<typename T,typename I>
		void TransposeSqrMatrix(T data[],I n)
		{
			for (I i= 0; i < n; ++i, data+=n+1 ) 
			{
				T *p1=data;
				T *p2=data;
				for(I j=i+1; j<n ; ++j)
				{
					++p1;
					p2+=n;
					std::swap(*p1,*p2);
				}
			}
/*
			size_t rows = size;
			for (size_t row = 0; row < rows; row++) 
			{
				T *r = img+row*size;
				for (size_t col = row + 1; col < rows; col++)
				{
					T *r2 = img+col*size;
					std::swap(r[col], r2[row]);
				}
			}
*/
		}

		/*!
		* \brief   ""    .
		*
		*         .
		* ,         \c func.\n
		*  Functor   : 
		* \code
		*	class func
		*	{
		*		//..
		*	public:
		*		void operator()(I j_first,I n,I step) {}
		*		enum { sm;};
		*	};
		* \endcode
		*     ,    .\n
		* ,  N-  (dim=N,    nn[k]   k, 0<=k<ndim) 
		*       ,    ...     
		*  ,  ,         
		*    .             
		*  . ,      ,      
		* (   ),     (  )      ,  
		*      , ""    
		*  M,    sm    M,    sm=1.
		*   ,         sm=1,    
		*              sm=2.\n
		*  \n operator()()         
		*  \c n,       \c step,        \c j_first. 
		*  \c step  \c j_first    ,   \c sm    
		*  \c ForEachDimensionDo1D. 
		*     \see IntTransf::Fourier::fftND(), IntTransf::Hartley::quasi_fhtND()
		*/
		template<size_t ndim,class Functor,typename I>
		void ForEachDimensionDo1D(Functor &func, I nn[])
		{
			// Compute total number of complex values.
			I ntot=ArrayElemProduct<ndim>(nn);
			I n_before=1;
			for(I k=0; k<ndim; ++k)
			{
				I n=nn[k];
				I n_to=n_before*n;
				I n_after=ntot/n_to;
				for(I j=0; n_after-- ; j+=n_to)
				{
					I cnt2=n_before;
					for(I j2=j*func.sm; cnt2-- ; j2+=func.sm) //sm - is step multiplier
						func(j2,n,n_before);
				}
				n_before=n_to;
			}
		}

		/*!
		 *       (  ) . 
		 */
		template<size_t ndim,bool re_im_at_sep_loc,typename I>
		I offset_to_zero_freq(I nn[])
		{
			size_t ofs=0;
			I n=1;
			for(int i=0; i < ndim; ++i)
			{
				n*=nn[i];
				ofs+=n/2;
			}
			return re_im_at_sep_loc? ofs : (ofs<<1);
		}

		/*!
		 * \brief   static void Vtl::NumAlg::IntTransf::Utils::proc_pos_neg_freq(I n[],ProcPair pp); 
		 *******************************************************************************************/
		namespace F
		{
			/*!
			 * \brief   static void proc_pos_neg_freq(I n[],ProcPair pp);   
			 *         (  ),   fromUpToDown  . 
			 *             .
			 *
			 *             ,  .. ndim>2.
             *******************************************************************************************/
			template<unsigned ndim,typename T,bool fromUpToDown>
			class ReflectHalfToOther
			{
				T* amp_;
			private:
				template<typename I>
				void set_amp(I plus_k,I minus_k,Loki::Int2Type<true> )
				{
					amp_[minus_k]=amp_[plus_k];
				}
				template<typename I>
				void set_amp(I plus_k,I minus_k,Loki::Int2Type<false> )
				{
					amp_[plus_k]=amp_[minus_k];
				}
			public:
				ReflectHalfToOther(T amp[]) 
					: amp_(amp)
				{}
				template<bool upper_half,typename I>
				void operator()(I plus_k,I minus_k,Loki::Int2Type<upper_half> )
				{
					set_amp(plus_k,minus_k,Loki::Int2Type<fromUpToDown>());
				}
				template<typename I>
				void operator()(I k,Loki::Int2Type<false>)
				{
					amp_[k]=0;
				}
				template<typename I>
				void operator()(I k,Loki::Int2Type<true>)
				{
					amp_[k]=0;
				}
			};

			/*!
			 * \brief   static void proc_pos_neg_freq(I n[],ProcPair pp);  .
			 *
			 */
			template<typename T>
			class Test
			{
				T* data_;
			public:
				//enum { step=2 };
				Test(T data[]) : data_(data) {}
				template<bool upper_half,typename I>
				void operator()(I plus_k,I minus_k,Loki::Int2Type<upper_half> )
				{
					data_[plus_k]+=0.2f;
					data_[minus_k]+=0.1f;
				}
//				template<bool upper_half,typename I>
//				void operator()(I k,Loki::Int2Type<upper_half> )
//				{
//					data_[k]=99.01f;
//				}
				template<typename I>
				void operator()(I k,Loki::Int2Type<true> )
				{
					data_[k]=99.01f;
				}
				template<typename I>
				void operator()(I k,Loki::Int2Type<false> )
				{
					data_[k]=98.01f;
				}
			};
		} //namespace ProcPair
		namespace Impl
		{
			template<size_t ndim,bool upper_half>
			struct proc_pos_neg_freq_
			{
				template<typename I,class ProcPair>
				static void job_c(I n[],I x1,I x2,ProcPair pp)
				{
					proc_pos_neg_freq_<ndim-1,upper_half>::job_single_left(n,x1,pp);
					for(I c2=n[ndim-1]/2-1; c2-- ;)
						proc_pos_neg_freq_<ndim-1,upper_half>::job_nc(n,x1,x2,pp);
					proc_pos_neg_freq_<ndim-1,false>::job_c(n,x1,x2,pp);
				}
				template<typename I,class ProcPair>
				static void job_single_left(I n[],I& x1,ProcPair pp)
				{
					I c=n[ndim-1];
					while( c-- )
						proc_pos_neg_freq_<ndim-1,upper_half>::job_single_left(n,x1,pp);
				}
				template<typename I,class ProcPair>
				static void job_single_right(I n[],I& x2,ProcPair pp)
				{
					I c=n[ndim-1];
					while( c-- )
						proc_pos_neg_freq_<ndim-1,upper_half>::job_single_right(n,x2,pp);
				}
				template<typename I,class ProcPair>
				static void job_nc(I n[],I& x1,I& x2,ProcPair pp)
				{
					proc_pos_neg_freq_<ndim-1,upper_half>::job_single_left(n,x1,pp);
					I c=n[ndim-1]-1;
					while( c-- )
						proc_pos_neg_freq_<ndim-1,upper_half>::job_nc(n,x1,x2,pp);
					proc_pos_neg_freq_<ndim-1,upper_half>::job_single_right(n,x2,pp);
				}
			};
			template<bool upper_half>
			struct proc_pos_neg_freq_<0,upper_half>
			{
				template<typename I,class ProcPair>
				static void job_c(I n[],I& x1,I& x2,ProcPair pp)
				{
					assert(x1==x2);
					pp(x2,x1,Loki::Int2Type<upper_half>());
				}
				template<typename I,class ProcPair>
				static void job_single_left(I n[],I& x1,ProcPair pp)
				{
					pp(x1++,Loki::Int2Type<upper_half>());
				}
				template<typename I,class ProcPair>
				static void job_single_right(I n[],I& x2,ProcPair pp)
				{
					pp(x2--,Loki::Int2Type<false>());
				}
				template<typename I,class ProcPair>
				static void job_nc(I n[],I& x1,I& x2,ProcPair pp)
				{
					pp(x2--,x1++,Loki::Int2Type<upper_half>());
				}
			};
		}
		/*!
		 * \brief        ,   
		 *                k  -k , 
		 *        ( -,           .).
		 *        
		 *          , 
		 *        \see namespace Vtl::NumAlg::IntTransf::Utils::F
		 *        \see namespace Vtl::NumAlg::IntTransf::Fourier::F
		 *        \see namespace Vtl::NumAlg::IntTransf::Hartley::F
		 *
		 *            , ndim>=1.
         *******************************************************************************************/
		template<size_t ndim,typename I,class ProcPair>
		static void proc_pos_neg_freq(I n[],ProcPair pp)
		{
			I x1=0;
			I x2=Utils::ArrayElemProduct<ndim>(n)-1;
			Impl::proc_pos_neg_freq_<ndim,true>::job_c(n,x1,x2,pp);
		}
	};
	/*! \brief   .     \ref Lit_NRC2 ,  \c four1 . 
	 *
	 *       ++,   -  ,   
	 *    (<tt> float, double, long double </tt>).\n
	 *      :\n
	 *    \f$ \begin{array}{cc} H(f) = \int\limits_{-\infty}^{+\infty} h(x) e^{ + 2 \pi i f x} dx   \\
	 *                          h(x) = \int\limits_{-\infty}^{+\infty} H(f) e^{ - 2 \pi i f x} df   \end{array} \f$ \n
     *
	 *
	 *******************************************************************************************/
	namespace Fourier
	{
		/*! \brief     , ..     . 
		 *    ,    .
		 ***************************************************************************************/
		template<typename T, typename I>
		void norm_ifft(T data[],I n,I step=1)
		{
			T multiplier=T(1.0/n);
			Utils::mult(data,n,multiplier,step);
		}
		/*! \brief  ,         re, im. n-    2.
		 ***************************************************************************************/
		template<typename T, typename I>
		void fft1D(T re[],T im[], I n)
		{
			fft1D<false>(re,im, n, 1);
		}
		/*! \brief  \b  ,         re, im. n-    2.
		 ***************************************************************************************/
		template<typename T, typename I>
		void ifft1D(T re[],T im[], I n)
		{
			fft1D<true>(re,im, n, 1);
		}
		/*! \brief  ,       e data 
		 *      (re, im). n-    2, 
		 *   data = 2*n.
		 ***************************************************************************************/
		template<typename T, typename I>
		void fft1D(T data[], I n)
		{
			fft1D<false>(data,data+1, n, 2);
		}
		/*! \brief  \b  ,       e data 
		 *      (re, im). n-    2, 
		 *   data = 2*n.
		 ***************************************************************************************/
		template<typename T, typename I>
		void ifft1D(T data[], I n)
		{
			fft1D<true>(data,data+1, n, 2);
		}
		/*! \brief  ,         re, im. \n
		 * nn -   2 , n[0] -     (, ..  ), 
		 * n[1] -    (, ..  ).       2.
		 ***************************************************************************************/
		template<typename T, typename I>
		void fft2D(T re[],T im[], I nn[])
		{
			fftND<2,false,true>(re,im, nn);
		}
		/*! \brief  ,         re, im. \n
		 * width -     (, ..  ), \n
		 * height -    (, ..  ).       2.
		 ***************************************************************************************/
		template<typename T, typename I>
		void fft2D(T re[],T im[], I width, I height)
		{
			I nn[]={width, height};
			fft2D(re,im,nn);
		}
		/*! \brief  \b  ,         re, im. \n
		 * nn -   2 , n[0] -     (, ..  ), \n
		 * n[1] -    (, ..  ).       2.
		 ***************************************************************************************/
		template<typename T, typename I>
		void ifft2D(T re[],T im[], I nn[])
		{
			fftND<2,true,true>(re,im, nn);
		}
		/*! \brief  \b  ,         re, im. \n
		 * width -     (, ..  ), \n
		 * height -    (, ..  ).       2.
		 ***************************************************************************************/
		template<typename T, typename I>
		void ifft2D(T re[],T im[], I width, I height)
		{
			I nn[]={width, height};
			ifft2D(re,im, nn);
		}
		/*! \brief  ,           (re,im). \n
		 * nn -   2 , n[0] -     (, ..  ), \n
		 * n[1] -    (, ..  ).       2.
		 ***************************************************************************************/
		template<typename T, typename I>
		void fft2D(T data[], I nn[])
		{
			fftND<2,false,false>(data,data+1, nn);
		}
		/*! \brief  ,           (re,im). \n
		 * width -     (, ..  ), \n
		 * height -    (, ..  ).       2.
		 ***************************************************************************************/
		template<typename T, typename I>
		void fft2D(T data[], I width, I height)
		{
			I nn[]={width, height};
			fft2D(data, nn);
		}
		/*! \brief  \b  ,           (re,im). 
		 * nn -   2 , n[0] -     (, ..  ), \n
		 * n[1] -    (, ..  ).       2.
		 ***************************************************************************************/
		template<typename T, typename I>
		void ifft2D(T data[], I nn[])
		{
			fftND<2,true,false>(data,data+1, nn);
		}
		/*! \brief  \b  ,           (re,im). 
		 * width -     (, ..  ), \n
		 * height -    (, ..  ).       2.
		 ***************************************************************************************/
		template<typename T, typename I>
		void ifft2D(T data[], I width, I height)
		{
			I nn[]={width, height};
			ifft2D(data, nn);
		}
		/*! \brief  ,         re, im. \n
		 * nn -   3 , n[0] -     (, ..  ), \n
		 * n[1] -    (, ..  ), \n 
		 * n[2] -    (.. ).  3     2.
		 ***************************************************************************************/
		template<typename T, typename I>
		void fft3D(T re[],T im[], I nn[])
		{
			fftND<3,false,true>(re,im, nn);
		}
		/*! \brief  ,         re, im. \n
		 * n_x  -     (, ..  ), \n
		 * n_y -    (, ..  ), \n 
		 * n_z -    (.. ).  3     2.
		 ***************************************************************************************/
		template<typename T, typename I>
		void fft3D(T re[],T im[], I n_x, I n_y,I n_z)
		{
			I nn[]={n_x,n_y,n_z};
			fft3D(re,im,nn);
		}
		/*! \brief  \b  ,         re, im. \n
		 * nn -   3 , n[0] -     (, ..  ), \n
		 * n[1] -    (, ..  ), \n 
		 * n[2] -    (.. ).  3     2.
		 ***************************************************************************************/
		template<typename T, typename I>
		void ifft3D(T re[],T im[], I nn[])
		{
			fftND<3,true,true>(re,im, nn);
		}
		/*! \brief  \b  ,         re, im. \n
		 * n_x -     (, ..  ), \n
		 * n_y -    (, ..  ), \n 
		 * n_z -    (.. ).  3     2.
		 ***************************************************************************************/
		template<typename T, typename I>
		void ifft3D(T re[],T im[], I n_x, I n_y,I n_z)
		{
			I nn[]={n_x,n_y,n_z};
			ifft3D(re,im,nn);
		}
		/*! \brief  ,           (re,im). \n
		 * nn -   3 , n[0] -     (, ..  ), \n
		 * n[1] -    (, ..  ), \n 
		 * n[2] -    (.. ).  3     2.
		 ***************************************************************************************/
		template<typename T, typename I>
		void fft3D(T data[], I nn[])
		{
			fftND<3,false,false>(data,data+1, nn);
		}
		/*! \brief  ,           (re,im). \n
		 * n_z -     (, ..  ), \n
		 * n_y -    (, ..  ), \n 
		 * n_z -    (.. ).  3     2.
		 ***************************************************************************************/
		template<typename T, typename I>
		void fft3D(T data[], I n_x, I n_y,I n_z)
		{
			I nn[]={n_x,n_y,n_z};
			fft3D(data,nn);
		}
		/*! \brief  \b  ,           (re,im). \n
		 * nn -   3 , n[0] -     (, ..  ), \n
		 * n[1] -    (, ..  ), \n 
		 * n[2] -    (.. ).  3     2.
		 ***************************************************************************************/
		template<typename T, typename I>
		void ifft3D(T data[], I nn[])
		{
			fftND<3,true,false>(data,data+1, nn);
		}
		/*! \brief  \b  ,           (re,im). \n
		 * n_x -     (, ..  ), \n
		 * n_y -    (, ..  ), \n 
		 * n_z -    (.. ).  3     2.
		 ***************************************************************************************/
		template<typename T, typename I>
		void ifft3D(T data[], I n_x, I n_y,I n_z)
		{
			I nn[]={n_x,n_y,n_z};
			ifft3D(data,nn);
		}
		/*! \brief   ,    (re,im)    step.  , 
		 *    (step=1    /   step=2  
		 *  )      (  step).
		 *  -  ,        .
		 *
		 * Replaces data[1..2*n] by its discrete Fourier transform, if \c 'reverse' is \c false; or replaces
		 * data[1..2*n] by n times its inverse discrete Fourier transform, if \c 'reverse' is \c true.
		 * data is a complex array of length n or, equivalently, a real array of length 2*n. n MUST
		 * be an integer power of 2 (this is not checked for!).
		 */
		template<bool is_reverse,typename T, typename I>
		void fft1D(T re[],T im[], I n, I step=1)
		{
			Utils::bit_reversal_complex(re,im, n,step);
			I nq=n*step;
			for(I m=2,mq=step ; m <= n ; m<<=1 )
			{
				I mq2=mq;//mq2=mq/2
				mq<<=1;
				Utils::Impl::TrigRecurrence<T,double,I> tr(is_reverse,m); 
				for(I j=0 ; j < mq2 ; j+=step )
				{
					for(I k=j ; k < nq ; k+=mq )
						tr.butterfly_fft(re,im,k,k+mq2);
					tr.next();
				}
			}
		}
		/*! \brief  ndim- ,     (re,im)  re_im_at_sep_loc=true 
		          ( im=re+1).  is_reverse= -   ,
		    - .

		   Replaces data by its ndim-dimensional discrete Fourier transform, if \c 'reverse' is \c false.
		   nn[1..ndim] is an integer array containing the lengths of each dimension (number of complex
		   values), which MUST all be powers of 2. data is a real array of length twice the product of
		   these lengths, in which the data are stored as in a multidimensional complex array: real and
		   imaginary parts of each element are in consecutive locations, and the rightmost index of the
		   array increases most rapidly as one proceeds along data. For a two-dimensional array, this is
		   equivalent to storing the array by rows. If \c 'reverse' is \c true, data is replaced by its inverse
		   transform times the product of the lengths of all dimensions.
		   
		    re_im_at_sep_loc=true     (re,im)     
		    ,    -         
		   ,    re=data, im=data+1,         
		         .
		    :
		       ndim=2.
		         ,   .
		    ndim=3,           , 
		      ndim=2.
		   ...
		    j  k-  0 <= k < ndim    0  nn[k].
		*/
		template<size_t ndim,bool is_reverse,bool re_im_at_sep_loc,typename T, typename I>
		void fftND(T re[],T im[], I nn[])
		{
			class func
			{
				T *re;
				T *im;
			public:
				func(T re_[],T im_[]) : re(re_),im(im_) {}
				void operator()(I j2,I n,I step)
				{
					Fourier::fft1D<is_reverse>(re+j2,im+j2, n, step);
				}
				enum { sm = re_im_at_sep_loc?1:2};//step multiplier 
			};
			func f(re,im);
			Utils::ForEachDimensionDo1D<ndim>(f, nn);
		}
		/*!
		 * \brief   static void Vtl::NumAlg::IntTransf::Utils::proc_pos_neg_freq(I n[],ProcPair pp); 
		 *******************************************************************************************/
		namespace F 
		{
			namespace Impl
			{
				template<typename T>
				void calc_F1multF2star_from_2_simult_real_fft(T& reQk,T& imQk,T& reQ_k,T& imQ_k)
				{
						T Re1= reQk + reQ_k;
						T Im1= imQk - imQ_k;
						T Re2= imQk + imQ_k;
						T Im2=-reQk + reQ_k;
						reQk = Re1*Re2+Im1*Im2;
						reQ_k=reQk;
						imQk = -Re1*Im2+Re2*Im1;
						imQ_k= -imQk;
				}
			}
			/*!
			 * \brief   static void proc_pos_neg_freq(I n[],ProcPair pp);    
			 *               I1,     - I2   
			 *         ,   proc_pos_neg_freq()       inplace
			 *         F1.F2* .
			 *
			 *             ,  .. ndim>2.
			 */
			template<bool re_im_at_sep_loc,typename T> class ClalcF1F2conj;
			template<typename T>
			class ClalcF1F2conj<true,T>
			{
				T* re_;
				T* im_;
			public:
				//enum { step=1 };
				ClalcF1F2conj(T re[],T im[]) :re_(re),im_(im) {}
				// j2 -> k , j1 -> -k
				template<bool upper_half,typename I>
				void operator()(I plus_k,I minus_k,Loki::Int2Type<upper_half> up_hlf)
				{
					Impl::calc_F1multF2star_from_2_simult_real_fft(re_[plus_k],im_[plus_k],re_[minus_k],im_[minus_k]);
				}
				template<bool upper_half,typename I>
				void operator()(I k,Loki::Int2Type<upper_half> )
				{
					re_[k]=0;
					im_[k]=0;
				}
			};
			/*!
			 * \brief   static void proc_pos_neg_freq(I n[],ProcPair pp);    
			 *               I1,     - I2   
			 *         ,   proc_pos_neg_freq()       inplace
			 *         F1.F2* .
			 *
			 *             ,  .. ndim>2.
			 */
			template<typename T>
			class ClalcF1F2conj<false,T>
			{
				T* data_;
			public:
				//enum { step=2 };
				ClalcF1F2conj(T data[]) : data_(data) {}
				template<bool upper_half,typename I>
				void operator()(I plus_k,I minus_k,Loki::Int2Type<upper_half> up_hlf)
				{
					plus_k<<=1;
					minus_k<<=1;
					Impl::calc_F1multF2star_from_2_simult_real_fft(data_[plus_k],data_[plus_k+1],data_[minus_k],data_[minus_k+1])
				}
				template<bool upper_half,typename I>
				void operator()(I k,Loki::Int2Type<upper_half> )
				{
					k<<=1;
					data_[k]=0;
					data_[k+1]=0;
				}
			};
			/*!
			 * \brief   static void proc_pos_neg_freq(I n[],ProcPair pp);    
			 *               I1,     - I2   
			 *         ,   proc_pos_neg_freq()       
			 *          I1  I2.
			 *
			 *             ,  .. ndim>2.
			 *
			 *  :\n
			 * \param first_second_or_both - 0 -  I1 \b  I2, 1 -  I1, 2 -  I2.
			 * \param layout  \n
			 *                 =0 -  ,     , \n
			 *                 =1 -   ,  I2    ,  I1 -  .\n
			 *                 =2 -          , 
			 *                        inplace . \n
			 *                     ,     
			 *                            ,      .   layou=1
			 *                        ""     .   layou=2  
			 *                          .
			 *
			 *    ,       
			 *   n-  .        f(x)
			 *  F(-k)=F*(k),  x,k -   n, ..   2D  () 
			 *       ,   
			 *  ,        
			 *   f1(x,y),     - f2(x,y).      
			 *           ,    .
			 *
			    \f$ f_1(x) \f$, \f$ f_2(x) \f$ - , 
			   \f$ x,k \in  \Re^n ,\space  f_1(x),f_2(x) \in  \Re \f$ . 
			       :
			   \f[  F_j(k) = \int\limits_{-\infty}^{+\infty} f_j(x) e^{ + 2 \pi i k x} d^n x    \f] 
			     : 
			   \f[  F_j(-k) = F_j^*(k) \f] .
			  
			            \f$  F_j(k) = Re(F_j(k)) + i Im(F_j(k)) \f$ . 
			      : \n
     			\f[ \left\lbrace   \begin{array}{cc}
					Re(F_j(-k))  = + Re(F_j(k))  \\ 
    				Im(F_j(-k))) = - Im(F_1(k))  
				\end{array} \right. \f] 

			     \f$ q(x)=f_1(x) + i f_2(x) \f$ . 
			       \f$ Q(k)=F_1(k) + i F_2(k) \f$ .
			         : 
     			\f[ \left\lbrace   \begin{array}{cc}
					Re(Q(k))=Re(F_1(k)) - Im(F_2(k)) \\ 
    				Im(Q(k))=Im(F_1(k)) + Re(F_2(k)) 
				\end{array} \right. \f] 

			    \f$ F_j(-k) = F_j^*(k) \quad  \forall j=1,2 \f$   \n
     			\f[ \left\lbrace {  \begin{array}{cc} 
					Re(Q(-k))=Re(F_1(-k)) - Im(F_2(-k)) = Re(F_1(k)) + Im(F_2(k)) \\ 
					Im(Q(-k))=Im(F_1(-k)) + Re(F_2(-k)) = - Im(F_1(k)) + Re(F_2(k)) 
				\end{array} } \right. \f] 

			  :\n
     			\f[ \left\lbrace  \begin{array}{cc} 
					Re(Q(k))  = + Re(F_1(k)) - Im(F_2(k)) \\ 
    				Im(Q(k))  = + Im(F_1(k)) + Re(F_2(k)) \\
					Re(Q(-k)) = + Re(F_1(k)) + Im(F_2(k)) \\ 
					Im(Q(-k)) = - Im(F_1(k)) + Re(F_2(k)) 
				\end{array} \right. \f] 

			  \n
     			\f[ 
					\left\lbrace {  \begin{array}{cc} 
					Re(F_1(k))  = {1 \over 2 } \space [ Re(Q(k)) + Re(Q(-k)) ]  \\ 
    				Im(F_1(k))  = {1 \over 2 } \space [ Im(Q(k)) - Im(Q(-k)) ]  \\
					Re(F_2(k))  = {1 \over 2 } \space [ Im(Q(k)) + Im(Q(-k)) ]  \\ 
					Im(F_2(k))  = {1 \over 2 } \space [ Re(Q(-k)) - Re(Q(k)) ]  \\
					Re(F_1(-k))  =  Re(F_1(k))  \\ 
					Im(F_1(-k))  = -Im(F_1(k))  \\
					Re(F_2(-k))  =  Re(F_2(k))  \\ 
					Im(F_2(-k))  = -Im(F_2(k))  \\
					Re(F_1(0)) = Re(Q(0)) \\
					Re(F_2(0)) = Im(Q(0)) \\
					Im(F_1(0))  = 0  \\
					Im(F_2(0))  = 0  
				\end{array} } \right. \f] 

			 ,   <b> </b>   n-   
			 <b>        </b> fftND<>,   
			   ,        . 
			            .

			=================================\n
			   ,        , 
			  ,      .    
			    2.    IntTransf::Utils::SwapQuadrants 
			\em (       !)     
			 (,  ...).  ""   ( )  
			   ().          
			  ,      1 ,  .
			             
			  .          
			    ( ).

			*******************************************************************************************/
			template<typename T,unsigned first_second_or_both=0,unsigned layout=0>
			class SplitEvenOdd
			{
				T* re;
				T* im;
				T* re1;
				T* im1;
				T* re2;
				T* im2;
			private:
				template<bool upper_half,typename I>
				void extract_(I plus_k,I minus_k,Loki::Int2Type<upper_half> up_hlf ,Loki::Int2Type<0> )
				{
					T r1  = 0.5*( re[plus_k]  + re[minus_k]);
    				T i1  = 0.5*( im[plus_k]  - im[minus_k]);
					T r2  = 0.5*( im[plus_k]  + im[minus_k]);
					T i2  = 0.5*( re[minus_k] - re[plus_k] );
					//   layout==0  layout==2, 
					//  layout==0  up_hlf    true - .  operator()
				 	set_1(r1,i1,plus_k,minus_k,up_hlf,Loki::Int2Type<layout&1>()); 
				 	set_2(r2,i2,plus_k,minus_k,up_hlf,Loki::Int2Type<layout&1>());
				}
				template<bool upper_half,typename I>
				void extract_(I plus_k,I minus_k,Loki::Int2Type<upper_half> up_hlf ,Loki::Int2Type<1> )
				{
					T r1  = 0.5*( re[plus_k]  + re[minus_k]);
    				T i1  = 0.5*( im[plus_k]  - im[minus_k]);
					//   layout==0  layout==2, 
					//  layout==0  up_hlf    true - .  operator()
				 	set_1(r1,i1,plus_k,minus_k,up_hlf,Loki::Int2Type<layout&1>());
				}
				template<bool upper_half,typename I>
				void extract_(I plus_k,I minus_k,Loki::Int2Type<upper_half> up_hlf ,Loki::Int2Type<2> )
				{
					T r2  = 0.5*( im[plus_k]  + im[minus_k]);
					T i2  = 0.5*( re[minus_k] - re[plus_k] );
					//   layout==0  layout==2, 
					//  layout==0  up_hlf    true - .  operator()
					set_2(r2,i2,plus_k,minus_k,up_hlf,Loki::Int2Type<layout&1>());
				}
				//layout==2
				template<typename I>
				void set_1(T r1,T i1,I plus_k,I minus_k,Loki::Int2Type<true> up_hlf ,Loki::Int2Type<0> )
				{
					re1[plus_k]= r1;
					im1[plus_k]= i1;
				}
				//layout==0 or 2
				template<typename I>
				void set_1(T r1,T i1,I plus_k,I minus_k,Loki::Int2Type<false> up_hlf ,Loki::Int2Type<0> )
				{
					re1[plus_k]= r1;
					im1[plus_k]= i1;
					re1[minus_k]= r1;
					im1[minus_k]=-i1;
				}
				//layout==2
				template<typename I>
				void set_2(T r2,T i2,I plus_k,I minus_k,Loki::Int2Type<true> up_hlf ,Loki::Int2Type<0> )
				{
					re2[plus_k]= r2;
					im2[plus_k]= i2;
				}
				//layout==0 or 2
				template<typename I>
				void set_2(T r2,T i2,I plus_k,I minus_k,Loki::Int2Type<false> up_hlf ,Loki::Int2Type<0> )
				{
					re2[plus_k]= r2;
					im2[plus_k]= i2;
					re2[minus_k]= r2;
					im2[minus_k]=-i2;
				}
				//layout==1
				template<typename I>
				void set_1(T r1,T i1,I plus_k,I minus_k,Loki::Int2Type<false> ,Loki::Int2Type<1> )
				{
					re1[plus_k]= r1;
					im1[plus_k]= i1;
				}
				//layout==1
				template<typename I>
				void set_2(T r2,T i2,I plus_k,I minus_k,Loki::Int2Type<false> ,Loki::Int2Type<1> )
				{
					re2[minus_k]= r2;
					im2[minus_k]=-i2;
				}
				template<typename I>
				void zero(I k,Loki::Int2Type<0> )
				{
					zero(k,Loki::Int2Type<1>() );
					zero(k,Loki::Int2Type<2>() );
				}
				template<typename I>
				void zero(I k,Loki::Int2Type<1> )
				{
					re1[k]=0;
					im1[k]=0;
				}
				template<typename I>
				void zero(I k,Loki::Int2Type<2> )
				{
					re2[k]=0;
					im2[k]=0;
				}
				template<typename I>
				void set_zero(I k,Loki::Int2Type<true> )
				{
				}
				template<typename I>
				void set_zero(I k,Loki::Int2Type<false> )
				{
					zero(k,Loki::Int2Type<both>() );
				}

			public:
				/// Some of destinations may may coinside with source
				SplitEvenOdd(T Q_re[],T Q_im[],T re_1[],T im_1[],T re_2[],T im_2[]) 
					: re(Q_re),im(Q_im),re1(re_1),im1(im_1),re2(re_2),im2(im_2)
				{
					LOKI_STATIC_CHECK((layout==0),This_constructor_form_used_ONLY_when_template_params_is__layout_is_0);
				}
				/// Some of destinations may may coinside with source
				SplitEvenOdd(T Q_re[],T Q_im[],T re_12[],T im_12[]) 
					: re(Q_re),im(Q_im),re1(re_12),im1(im_12),re2(re_12),im2(im_12)
				{
					LOKI_STATIC_CHECK((layout==1),This_constructor_form_used_ONLY_when_template_params_is__layout_is_1);
				}
				/// Destinations CAN NOT coinside with source! Destinations is half size of source!
				template<unsigned ndim>
				SplitEvenOdd(T Q_re[],T Q_im[],T re_1[],T im_1[],T re_2[],T im_2[],size_t n[]) 
					: re(Q_re),im(Q_im)
					,re1(re_1-ArrayElemProduct<ndim>(n)/2)
					,im1(im_1-ArrayElemProduct<ndim>(n)/2)
					,re2(re_2-ArrayElemProduct<ndim>(n)/2)
					,im2(im_2-ArrayElemProduct<ndim>(n)/2)
				{
					LOKI_STATIC_CHECK((layout==2),This_constructor_form_used_ONLY_when_template_params_is__layout_is_2);
					LOKI_STATIC_CHECK(((Q_re!=re_1)&&(Q_re!=im_1)&&(Q_re!=re_2)&&(Q_re!=im_2)&&(Q_im!=re_1)&&(Q_im!=im_1)&&(Q_im!=re_2)&&(Q_im!=im_2)),This_constructor_form_used_ONLY_when_template_params_is__layout_is_2__it_cant_do_inplace);
				}
				template<unsigned ndim>
				SplitEvenOdd(T Q_re[],T Q_im[],T re_dest[],T im_dest[],size_t n[]) 
					: re(Q_re),im(Q_im)
					,re1(re_dest-ArrayElemProduct<ndim>(n)/2)
					,im1(im_dest-ArrayElemProduct<ndim>(n)/2)
					,re2(re_dest)
					,im2(im_dest)
				{
					LOKI_STATIC_CHECK((layout==2),This_constructor_form_used_ONLY_when_template_params_is__layout_is_2);
					LOKI_STATIC_CHECK(((Q_re!=re_dest)&&(Q_re!=im_dest)&&(Q_im!=re_dest)&&(Q_im!=im_dest)),This_constructor_form_used_ONLY_when_template_params_is__layout_is_2__it_cant_do_inplace);
				}
				template<bool upper_half,typename I>
				void operator()(I plus_k,I minus_k,Loki::Int2Type<upper_half> up_hlf)
				{
					//.. up_hlf==true -    layout==2,     
					//    .    up_hlf=false
					//   .
					extract_(plus_k,minus_k,Loki::Int2Type<(up_hlf&&(layout==2))>(),Loki::Int2Type<both>() );
				}
				template<bool upper_half,typename I>
				void operator()(I k,Loki::Int2Type<upper_half> up_hlf)
				{
					//.. up_hlf==true -    layout==2,     
					//    .    up_hlf=false
					//   .
					set_zero(k,Loki::Int2Type<(up_hlf&&(layout==2))>() );
				}
			};
		}//namespace F 
	}//namespace Fourier
	/*! \brief    \ref Lit_Brace. 
	 *  
	 *    -  ,   
	 *    (<tt> float, double, long double </tt>).\n
	 *      :\n
	 *    \f$ \begin{array}{cc} H(f) = \int\limits_{-\infty}^{+\infty} h(x) cas( 2 \pi f x) dx   \\
	 *                          h(x) = \int\limits_{-\infty}^{+\infty} H(f) cas(  2 \pi f x) df   \end{array} \f$ \n
	 *         , \f$ cas(x) \equiv cos(x)+sin(x) \f$
	 *     ,       N,  N - 
	 *   (   2),    .
	 *
     *    h(x)  ,    ,      , 
	 *     .          :\n
	 *    \f$ \begin{array}{cc} Re(F(k)) = (H(k)+H(-k))/2   \\
	 *                          Im(F(k)) = (H(k)-H(-k))/2   \\
	 *                          F(k) = F^{*}(-k)   \\
	 *                          |F(k)|^2 = {1 \over 2 } \space [ H(k)^2 + H(-k)^2 ]   \end{array} \f$ \n
	 *  !           ,  
	 *  IntTransf::Fourier
	 *
	 *******************************************************************************************/
	namespace Hartley
	{
		namespace Impl
		{
			template<typename T,typename I>
			static void dif_sum(T *data,I i,I j)
			{
				T d0 = data[i] - data[j];
				data[i] += data[j];
				data[j] = d0;
			}
		};
		/*-----------------------------------------------------*/
		//!      data.   n*log(n).
		template<typename T,typename I>
		void fht1D_(T *data, I n,I step)
		{
			if (n <= 0)	return;
			Utils::bit_reversal(data,n,step);

			//    2.
			const I ns=n*step;
			for (I i = 0,dd=step<<1 ; i < ns; i += dd)
				Impl::dif_sum(data,i,i+step);


			for (I blocksize = 4; blocksize <= n; blocksize <<= 1)
			{
				//int blockcount = n/blocksize;//    data
				const I bs=blocksize*step; 
				const I blocksize_2 = bs>>1;
				const I blocksize_4 = blocksize_2>>1;

				for (I i = 0; i < ns; i += bs)
				{
					I j = i+blocksize_2;
					Impl::dif_sum(data,i,j);
					Impl::dif_sum(data,i + blocksize_4,j + blocksize_4);
				}

				if (blocksize == 4) continue;

				Utils::Impl::TrigRecurrence<T,double,I> tr(false,blocksize);
				tr.next();
				for (I i = step ; i < blocksize_4; i+=step )
				{
					I j=blocksize_2-i;
					for (T *even=data; even<data+ns; even+=bs)
						tr.butterfly_Hartley(even,even+blocksize_2,i,j);

					tr.next();
				}
			}
		}
		/*!
		 * \brief   static void Vtl::NumAlg::IntTransf::Utils::proc_pos_neg_freq(I n[],ProcPair pp); 
		 *******************************************************************************************/
		namespace F 
		{
			template<typename T>
			class Hartley2Modul
			{
				T* data_in_;
				T* data_out_;
			public:
				//enum { step=2 };
				Hartley2Modul(T data_in[],T data_out[]) : data_in_(data_in),data_out_(data_out)  {}
				template<typename I>
				void operator()(I plus_k,I minus_k)
				{
					T s=std::sqrt(IntTransf::Utils::Sqr(data_in_[plus_k])+IntTransf::Utils::Sqr(data_in_[minus_k]));
					data_out_[plus_k]=s;
					data_out_[minus_k]=s;
				}
				template<typename I>
				void operator()(I plus_k)
				{
					data_out_[plus_k]=std::abs(data_in_[plus_k]);
				}
			};
			template<typename T>
			class Hartley2Re
			{
				T* data_in_;
				T* data_out_;
			public:
				//enum { step=2 };
				Hartley2Re(T data_in[],T data_out[]) : data_in_(data_in),data_out_(data_out)  {}
				template<bool upper_half,typename I>
				void operator()(I plus_k,I minus_k,Loki::Int2Type<upper_half> up_hlf)
				{
					T s=(data_in_[plus_k]+data_in_[minus_k])/2;
					data_out_[plus_k]=s;
					data_out_[minus_k]=s;
				}
				template<bool upper_half,typename I>
				void operator()(I k,Loki::Int2Type<upper_half> )
				{
					data_out_[k]=data_in_[k];
				}
			};
			template<typename T>
			class Hartley2Im
			{
				T* data_in_;
				T* data_out_;
			public:
				//enum { step=2 };
				Hartley2Im(T data_in[],T data_out[]) : data_in_(data_in),data_out_(data_out)  {}
				template<bool upper_half,typename I>
				void operator()(I plus_k,I minus_k,Loki::Int2Type<upper_half> up_hlf)
				{
					T s=(data_in_[plus_k]-data_in_[minus_k])/2;
					data_out_[plus_k]=s;
					data_out_[minus_k]=-s;
				}
				template<bool upper_half,typename I>
				void operator()(I k,Loki::Int2Type<upper_half> )
				{
					data_out_[k]=T(0);
				}
			};
			/*!
			 * \brief   static void proc_pos_neg_freq(I n[],ProcPair pp);    
			 *           (   ). 
			 *
			 *              
			 *         ,         ( ), 
			 *            ,  -    
			 *        . \see Hartley2Phase
			 *
			 *             ,  .. ndim>2.
			 */
			template<typename T,bool amp_in_up_half_plane,bool only_half_amp=false>
			class Hartley2Amp
			{
				T* hartley_;
				T* amp_;
			private:
				template<typename I>
				void set_amp(I plus_k,I minus_k,T amplitude)
				{
					set_amp(plus_k,minus_k,amplitude,Loki::Int2Type<amp_in_up_half_plane>());
				}//amp_in_up_half_plane
				template<typename I>
				void set_amp(I plus_k,I minus_k,T amplitude,Loki::Int2Type<true> )
				{
					amp_[minus_k]=amplitude;
				}
				template<typename I>
				void set_amp(I plus_k,I minus_k,T amplitude,Loki::Int2Type<false> )
				{
					amp_[plus_k]=amplitude;
				}
			public:
				Hartley2Amp(T hartley[],T amp[]) 
					: hartley_(hartley), amp_(amp)
				{}
				template<bool upper_half,typename I>
				void operator()(I plus_k,I minus_k,Loki::Int2Type<upper_half> up_hlf)
				{
					T a=Vtl::sqr(hartley_[plus_k])+Vtl::sqr(hartley_[minus_k]);
					a=std::sqrt(a/2);
					set_amp(plus_k,minus_k,a);
				}
				template<typename I>
				void operator()(I k,Loki::Int2Type<false>)
				{
					amp_[k]=0;
				}
				template<typename I>
				void operator()(I k,Loki::Int2Type<true>)
				{
				}
			};
			/*!
			 * \brief   static void proc_pos_neg_freq(I n[],ProcPair pp);    
			 *           (   ). 
			 *
			 *           
			 *         ,       ,    .
			 *               ( ), 
			 *             - .      amp. 
			 *
			 *             ,  .. ndim>2.
			 */
			template<unsigned ndim,typename T>
			class Hartley2AmpDown
			{
				T* hartley_;
				T* amp_;
			private:
				template<typename I>
				void set_amp(I plus_k,I minus_k,T amplitude,Loki::Int2Type<true> )
				{
					amp_[plus_k]=amplitude;
				}
				template<typename I>
				void set_amp(I plus_k,I minus_k,T amplitude,Loki::Int2Type<false> )
				{
					amp_[plus_k]=amplitude;
					amp_[minus_k]=amplitude;
				}
			public:
				Hartley2AmpDown(T hartley[],T amp[],size_t n[]) 
					: hartley_(hartley), amp_(amp-ArrayElemProduct<ndim>(n)/2)
				{}
				template<bool upper_half,typename I>
				void operator()(I plus_k,I minus_k,Loki::Int2Type<upper_half> up_hlf)
				{
					T a=Vtl::sqr(hartley_[plus_k])+Vtl::sqr(hartley_[minus_k]);
					a=std::sqrt(a/2);
					set_amp(plus_k,minus_k,a,up_hlf);
				}
				template<typename I>
				void operator()(I k,Loki::Int2Type<false>)
				{
					amp_[k]=0;
				}
				template<typename I>
				void operator()(I k,Loki::Int2Type<true>)
				{
				}
			};
			/*!
			 * \brief   static void proc_pos_neg_freq(I n[],ProcPair pp);    
			 *           (   )  ,   
			 *             . 
			 *
			 *                 .       
			 *         ,         ( ), 
			 *            ,  -    
			 *        . \see Hartley2Phase
			 *
			 *             ,  .. ndim>2.
			 */
			template<typename T,bool amp_in_up_half_plane,bool only_half_amp=false>
			class Hartley2PhaseAmp
			{
				T* hartley_;
				T* amp_;
				T* phase_;
			private:
				template<typename I>
				void set_amp(I plus_k,I minus_k,T amplitude)
				{
					set_amp(plus_k,minus_k,amplitude,Loki::Int2Type<amp_in_up_half_plane>());
				}//amp_in_up_half_plane
				template<typename I>
				void set_amp(I plus_k,I minus_k,T amplitude,Loki::Int2Type<true> )
				{
					amp_[minus_k]=amplitude;
				}
				template<typename I>
				void set_amp(I plus_k,I minus_k,T amplitude,Loki::Int2Type<false> )
				{
					amp_[plus_k]=amplitude;
				}
			public:
				Hartley2PhaseAmp(T hartley[],T amp[],T phase[]) 
					: hartley_(hartley), amp_(amp), phase_(phase)
				{}
				Hartley2PhaseAmp(T hartley[],T amp[],T phase[],unsigned ndim,size_t n[]) 
					: hartley_(hartley), amp_(only_half_amp?amp-ArrayElemProduct<ndim>(n)/2:amp), phase_(phase)
				{}
				Hartley2PhaseAmp(T hartley_phase[],T amp[]) 
					: hartley_(hartley_phase), amp_(amp), phase_(hartley_phase)
				{}
				template<unsigned ndim>
				Hartley2PhaseAmp(T hartley_phase[],T amp[],size_t n[]) 
					: hartley_(hartley_phase), amp_(only_half_amp?amp-ArrayElemProduct<ndim>(n)/2:amp), phase_(hartley_phase)
				{}
				template<bool upper_half,typename I>
				void operator()(I plus_k,I minus_k,Loki::Int2Type<upper_half> up_hlf)
				{
					T a=Vtl::sqr(hartley_[plus_k])+Vtl::sqr(hartley_[minus_k]);
					a=std::sqrt(a/2);
					set_amp(plus_k,minus_k,a);
					if(a>0)
					{
						phase_[plus_k]=hartley_[plus_k]/a;
						phase_[minus_k]=hartley_[minus_k]/a;
					}
					else
					{
						phase_[plus_k]=0;
						phase_[minus_k]=0;
					}
				}
				template<typename I>
				void operator()(I k,Loki::Int2Type<false>)
				{
					amp_[k]=0;
					phase_[k]=0;
				}
				template<typename I>
				void operator()(I k,Loki::Int2Type<true>)
				{
					phase_[k]=0;
				}
			};
			/*!
			 * \brief   static void proc_pos_neg_freq(I n[],ProcPair pp);    
			 *           (   )  ,   
			 *             . 
			 *
			 *                 .    
			 *         ,       ,   .
			 *               ( ), 
			 *             - .      amp. 
			 *
			 *             ,  .. ndim>2.
			 */
			template<unsigned ndim,typename T>
			class Hartley2PhaseAmpDown
			{
				T* hartley_;
				T* amp_;
				T* phase_;
			private:
				template<typename I>
				void set_amp(I plus_k,I minus_k,T amplitude,Loki::Int2Type<true> )
				{
					amp_[plus_k]=amplitude;
				}
				template<typename I>
				void set_amp(I plus_k,I minus_k,T amplitude,Loki::Int2Type<false> )
				{
					amp_[plus_k]=amplitude;
					amp_[minus_k]=amplitude;
				}
			public:
				Hartley2PhaseAmpDown(T hartley[],T amp[],T phase[],size_t n[]) 
					: hartley_(hartley), amp_(amp-ArrayElemProduct<ndim>(n)/2), phase_(phase)
				{}
				Hartley2PhaseAmpDown(T hartley_phase[],T amp[],size_t n[]) 
					: hartley_(hartley_phase), amp_(amp-Utils::ArrayElemProduct<ndim>(n)/2), phase_(hartley_phase)
				{}
				template<bool upper_half,typename I>
				void operator()(I plus_k,I minus_k,Loki::Int2Type<upper_half> up_hlf)
				{
					T a=Vtl::sqr(hartley_[plus_k])+Vtl::sqr(hartley_[minus_k]);
					a=std::sqrt(a/2);
					set_amp(plus_k,minus_k,a,up_hlf);
					if(a>0)
					{
						phase_[plus_k]=hartley_[plus_k]/a;
						phase_[minus_k]=hartley_[minus_k]/a;
					}
					else
					{
						phase_[plus_k]=0;
						phase_[minus_k]=0;
					}
				}
				template<typename I>
				void operator()(I k,Loki::Int2Type<false>)
				{
					amp_[k]=0;
					phase_[k]=0;
				}
				template<typename I>
				void operator()(I k,Loki::Int2Type<true>)
				{
					phase_[k]=0;
				}
			};
			/*!
			 * \brief   static void proc_pos_neg_freq(I n[],ProcPair pp);    
			 *          ,         .
			 *
			 *                 . \see Hartley2PhaseAmp
			 *
			 *             ,  .. ndim>2.
			 */
			template<typename T>
			class Hartley2Phase
			{
				T* hartley_;
				T* phase_;
			public:
				Hartley2Phase(T hartley[],T phase[]) 
					: hartley_(hartley), phase_(phase)
				{}
				Hartley2Phase(T hartley_phase[]) 
					: hartley_(hartley_phase), phase_(hartley_phase)
				{}
				template<bool upper_half,typename I>
				void operator()(I plus_k,I minus_k,Loki::Int2Type<upper_half> up_hlf)
				{
					T a=Vtl::sqr(hartley_[plus_k])+Vtl::sqr(hartley_[minus_k]);
					if(a>0)
					{
						a=std::sqrt(a/2);
						phase_[plus_k]=hartley_[plus_k]/a;
						phase_[minus_k]=hartley_[minus_k]/a;
					}
					else
					{
						phase_[plus_k]=0;
						phase_[minus_k]=0;
					}
				}
				template<bool upper_half,typename I>
				void operator()(I k,Loki::Int2Type<upper_half> )
				{
					phase_[k]=0;
				}
			};
			/*!
			 * \brief   static void proc_pos_neg_freq(I n[],ProcPair pp);   
			 *               F1.F2*   .
			 *                .
			 *             ,  .. ndim>2.
			 *
			 *              F1.F2* .    ,  
			 *         F1  F2   , ..   ,    
			 *          ,  ,       (dx,dy),
			 *              exp(2*pi*i*(kx*dx+ky*dy)).  
			 *              -       .
			 *
			 *           ,     
			 *        H1(x,y)H2(x,y)+H1(-x,-y)H2(-x,-y)+H1(-x,-y)H2(x,y)-H1(x,y)H2(-x,-y),  H1(.), H2(.) - 
			 *             ().       
			 *         ,      cas(2*pi*kx*dx+ky*dy)).  
			 *               -    
			 *         .
			 *
			 *         
			 *        \f$ f_2(\vec r)=f_1(\vec r - \vec r_0 )  \f$ \n
			 *        \f$ H(\vec u)=  \int\limits_{-\infty}^{+\infty} f(\vec r) cas( 2 \pi \vec u \vec r) d^n \vec r    \f$ \n
			 *        \f$ H_2(\vec u) = H_1(\vec u) cos( 2 \pi \vec u \vec r_0 ) +  H_1(- \vec u) sin( 2 \pi \vec u \vec r_0 )  \f$ \n
			 *        \f$ H_2(- \vec u) = H_1(- \vec u) cos( 2 \pi \vec u \vec r_0 ) -  H_1(\vec u) sin( 2 \pi \vec u \vec r_0 )  \f$ \n
			 *
			 *        :
			 *        \f$ \begin{array}{cc} 
			 *            H_1(  \vec u)H_2(  \vec u) =  H_1(  \vec u)H_1(  \vec u) cos( 2 \pi \vec u \vec r_0 ) + H_1(  \vec u)H_1(- \vec u) sin( 2 \pi \vec u \vec r_0 )   \\
			 *          + H_1(- \vec u)H_2(  \vec u) =  H_1(  \vec u)H_1(- \vec u) cos( 2 \pi \vec u \vec r_0 ) + H_1(- \vec u)H_1(- \vec u) sin( 2 \pi \vec u \vec r_0 )   \\
			 *          - H_1(  \vec u)H_2(- \vec u) = -H_1(  \vec u)H_1(- \vec u) cos( 2 \pi \vec u \vec r_0 ) + H_1(  \vec u)H_1(  \vec u) sin( 2 \pi \vec u \vec r_0 )   \\
			 *          + H_1(- \vec u)H_2(- \vec u) =  H_1(- \vec u)H_1(- \vec u) cos( 2 \pi \vec u \vec r_0 ) - H_1(  \vec u)H_1(- \vec u) sin( 2 \pi \vec u \vec r_0 )   \\
			 *          = [H_1^2( \vec u)+H_1^2(- \vec u)] cas( 2 \pi \vec u \vec r_0 )
			 *        \end{array} \f$
			 *
			 */
			template<typename T>
			class PhaseCorr
			{
				T* hartley1_;
				T* hartley2_;
				T* corr_;
			public:
				PhaseCorr(T hartley1[],T hartley2[],T corr[]) 
					: hartley1_(hartley1), hartley2_(hartley2), corr_(corr)
				{}
				template<bool upper_half,typename I>
				void operator()(I plus_k,I minus_k,Loki::Int2Type<upper_half> up_hlf)
				{
					T v1=hartley1_[plus_k]*hartley2_[plus_k]+hartley1_[minus_k]*hartley2_[minus_k];
					T v2=hartley1_[minus_k]*hartley2_[plus_k]-hartley1_[plus_k]*hartley2_[minus_k];
					corr_[plus_k]=(v1+v2)/2;
					corr_[minus_k]=(v1-v2)/2;
				}
				template<bool upper_half,typename I>
				void operator()(I k,Loki::Int2Type<upper_half> )
				{
					corr_[k]=0;
				}
			};
		}//namespace F 
		//!        .
		// hartley  real  
		template<int ndim,typename T,typename I>
		void GetModul(T *hartley, I n[], T *real)
		{
			F::Hartley2Modul<T> pp(hartley,real);
			IntTransf::Utils::proc_pos_neg_freq<ndim>(n,pp);
		}
		//!        .
		// hartley  real  
		template<int ndim,typename T,typename I>
		void GetReal(T *hartley, I n[], T *real)
		{
			F::Hartley2Re<T> pp(hartley,real);
			IntTransf::Utils::proc_pos_neg_freq<ndim>(n,pp);
		}
		template<typename T,typename I>
		void GetReal_1D(T *hartley, I n, T *real,I step=1,I step_real=1)
		{
			I sz2=n>>1;
			I last_h=n*step;
			I last_o=n*step_real;
			if(n==2)
			{
				real[step_real] = hartley[step];
			}
			else
			{
				for (I j=step,i =step_real ; sz2-- ; i+=step_real, j+=step)
				{
					T re = (hartley[j] + hartley[last_h-j])/2;
					real[i] = re;
					real[last_o-i] = re;
				}
			}
			real[0] = hartley[0];
		}
		//!        .
		// hartley  imag  
		template<int ndim,typename T,typename I>
		void GetImag(T *hartley, I n[], T *imag)
		{
			F::Hartley2Im<T> pp(hartley,imag);
			IntTransf::Utils::proc_pos_neg_freq<ndim>(n,pp);
		}
		template<typename T,typename I>
		void GetImag_1D(T *hartley, I n, T *imag,I step=1,I step_imag=1)
		{
			I sz2=n>>1;
			I last_h=n*step;
			I last_o=n*step_imag;
			if(n==2)
			{
				imag[step_imag] = 0;
			}
			else
			{
				for (I j=step,i =step_imag ; sz2-- ; i+=step_imag, j+=step)
				{
					T im = (hartley[j] - hartley[last_h-j])/2;
					imag[i] = im;
					imag[last_o-i] = -im;
				}
			}
			imag[0] = 0;
		}
		//!     .
		template<typename T,typename I>
		void AddQuadrants2D(T data[],I w,I h)
		{
			assert(w == h);
//			std::cout << "AddQuadrants2D===" <<std::endl;
			if(h<2 ||w<2)
				return;
			//      
			T *row1 = data+w;
			T *row2 = data+w*(h-1);
			for (I h2=h>>1,w2=w>>1;  --h2; row1+=w,row2-=w)
			{
				for (I x = 1; x < w2; x++)
				{
					T d1 = row1[x];
					T d2 = row1[w - x];
					T d3 = row2[x];
					T d4 = row2[w - x];

					row1[x]     =	T(0.5)*(d1 + d2 + d3 - d4);
					row1[w - x] =	T(0.5)*(d4 + d1 + d2 - d3);
					row2[x]     =	T(0.5)*(d3 + d4 + d1 - d2);
					row2[w - x] =	T(0.5)*(d2 + d3 + d4 - d1);
				}
			}
		}

		//!      data.   n*log(n).
		template<typename T,typename I>
		void fht2D_square(T data[],I w,I h)
		{
			assert(w == h);

			I n[]={w,h};
			quasi_fht2D(data,w);
			AddQuadrants2D(data,w,h);
		}
		//!    ,      data.
		template<typename T,typename I>
		void quasi_fht2D(T data[],I n)
		{
			//int row;
			{// 1D     
				T *d=data;
				for (I i = n; i-- ; d+=n )
					fht1D_(d, n,I(1));
			}
			// 
			Utils::TransposeSqrMatrix(data,n);
			// 1D     
			{// 1D     
				T *d=data;
				for (I i = n; i-- ; d+=n )
					fht1D_(d, n,I(1));
			}
			// 
			Utils::TransposeSqrMatrix(data,n);
		}

		
		/*!
		 *     ,    ,
		 *  .      
		 *    ,   2D -   \see AddQuadrants2D
		 */
		template<size_t ndim,typename T,typename I>
		void quasi_fhtND(T data[], I nn[])
		{
			class func
			{
				T *data;
			public:
				func(T data_[]) : data(data_){}
				void operator()(I j2,I n,I step)
				{
					fht1D_(data+j2, n, step);
				}
				enum { sm=1 };//step multiplier is 1 - data is real, one sample follows another
				
			};
			func f(data);
			Utils::ForEachDimensionDo1D<ndim>(f, nn);
		}
/*
		template<typename T,typename I>
		void quasi_fhtND(T data[] unsigned long nn[], int ndim)
		{
			const I sm=1;
			I ntot=1;
			for(int idim=0;idim <ndim;idim++) // Compute total number of complex values.
				ntot *= nn[idim];
			I n_before=1;
			for(I k=0; k<ndim; ++k)
			{
				I n=nn[k];
				n_to=n_before*n;
				I n_after=ntot/n_to;
				for(I j=0; n_after-- ; j+=n_to)
				{
					I cnt2=n_before;
					for(I j2=j*sm; cnt2-- ; j2+=sm)
						fht1D_(data+j2, n, n_before);
				}
				n_before=n_to;
			}
		}
*/
		//!    
		template<typename T, typename I>
		void fht1D(T data[], I n)
		{
//			Utils::SwapArrayHalfs<false>(data, n, I(1) );
			fht1D_(data, n, I(1) );
//			Utils::SwapArrayHalfs<false>(data, n, I(1) );
		}
		template<typename T, typename I>
		void ifht1D(T data[], I n)
		{
//			Utils::SwapArrayHalfs<false>(data, n, I(1) );
			fht1D_(data, n, I(1) );
//			Utils::SwapArrayHalfs<false>(data, n, I(1) );
		}
		template<typename T, typename I>
		void fht2D(T data[], I width, I height)
		{
			fht2D_square(data,width,height);
		}
		template<typename T, typename I>
		void ifht2D(T data[], I width, I height)
		{
			fht2D_square(data,width,height);
		}
	};
	/*
		//  
	template<typename T> struct CHartley
	{
		static void bit_reversal(T *array, int size)
		{
			//    .
			//   i 
			//     j,
			//   i    j,    
			//.    (n)  8.    3 (log(n)) .
			// i = 0 = 000 <=> j = 0 = 000
			// i = 1 = 001 <=> j = 4 = 100
			// i = 2 = 010 <=> j = 2 = 010
			// i = 3 = 011 <=> j = 6 = 110
			// i = 4 = 100 <=> j = 1 = 001
			// i = 5 = 101 <=> j = 5 = 101
			// i = 6 = 110 <=> j = 3 = 011
			// i = 7 = 111 <=> j = 7 = 111

			if (size <= 2) return;
			int i, l;//l = log(size)
			l = 0;
			while ((1 << l) < size) l++;
			for (i = 1; i < size - 1; i++)
			{
				int j, k, m;
				j = 0;
				k = i;
				m = l;
				while (m--)
				{
					j <<= 1;
					j |= (k & 1);
					k >>= 1;
				}

				if (j > i)
					std::swap(array[i], array[j]);
			}
		}
		//!      data.   n*log(n).
		static void Do1DInternal(T *data, int n)
		{
			if (n <= 0)	return;
			bit_reversal(data, n);

			int blocksize, blocksize_4, blocksize_2;// 
			int blockcount;//    data
			int i, j, k, l;
			T d0, d1, c, s, c0, s0;
			T *even, *odd;

			//    2.
			for (i = 0, j = 1; i < n; i += 2, j += 2)
			{
				d0 = data[i] - data[j];
				data[i] += data[j];
				data[j] = d0;
			}

			for (blocksize = 4; blocksize <= n; blocksize <<= 1)
			{
				blockcount = n/blocksize;
				blocksize_4 = blocksize/4;
				blocksize_2 = blocksize/2;

				for (i = 0, j = blocksize_2; i < n; i += blocksize, j += blocksize)
				{
					d0 = data[i] - data[j];
					data[i] += data[j];
					data[j] = d0;
					k = i + blocksize_4;
					l = j + blocksize_4;
					d0 = data[k] - data[l];
					data[k] += data[l];
					data[l] = d0;
				}

				if (blocksize == 4) continue;

				c = (T)cos(2*M_PI/blocksize);
				s = (T)sqrt(1-c*c);
				c0 = c;
				s0 = s;
				for (i = 1, j=blocksize_2-1; i < blocksize_4; i++, j--)
				{
					c = (T)cos(2*M_PI*i/blocksize);
					s = (T)sqrt(1-c*c);

					for (even=data, odd=data + blocksize_2; even<data+n; even+=blocksize, odd+=blocksize)
					{
						d0 = c*odd[i] + s*odd[j];
						d1 = c*odd[j] - s*odd[i];
						odd[i] = even[i] - d0;
						odd[j] = even[j] + d1;
						even[i] += d0;
						even[j] += - d1;
					}

					d0 = c*c0 - s*s0;
					s = s*c0 + c*s0;
					c = d0;
				}
			}
		}
		//!   .
		static void ImageTranspose(T* img,size_t size)
		{
			size_t rows = size;
			for (size_t row = 0; row < rows; row++) 
			{
				T *r = img+row*size;
				for (size_t col = row + 1; col < rows; col++)
				{
					T *r2 = img+col*size;
					std::swap(r[col], r2[row]);
				}
			}
		}
		//!      
		static void TransposeQuarters(T* img,size_t size)
		{
			size_t w = size;
			size_t h = size;

			//     
			for (size_t  i = 0; i < h/2; ++i)
			{
				T* row1 = img+i;
				T* row2 = img+w*(i+h/2);
				for (size_t j = 0; j < w/2; ++j)
				{
					//     
					std::swap(row1[j], row2[j+w/2]);
					//     
					std::swap(row1[j+w/2], row2[j]);
				}
			}
		}

		//!    ,     img.
		static void Do1DRowsAndCols(T* img,size_t size)
		{
			// 1D     
			for (size_t row = 0; row < size; ++row)
				Do1DInternal(img+size*row, size);

			// 
			ImageTranspose(img,size);

			// 1D     
			for (size_t row = 0; row < size; ++row)
				Do1DInternal(img+size*row, size);

			// 
			ImageTranspose(img,size);
		}
		//!     .
		static void AddQuadrants(T* img,size_t size)
		{
			size_t rows = size;
			//      
			for(size_t row = 1; row < rows/2; ++row)
			{
				T* row1 = img+size*row;
				T* row2 = img+size*(rows - row);
				for(size_t col = 1; col < rows/2; col++)
				{
					T d1 = row1[col];
					T d2 = row1[rows - col];
					T d3 = row2[col];
					T d4 = row2[rows - col];

					row1[col] =			(T)0.5*(d1 + d2 + d3 - d4);
					row1[rows - col] =	(T)0.5*(d4 + d1 + d2 - d3);
					row2[col] =			(T)0.5*(d3 + d4 + d1 - d2);
					row2[rows - col] =	(T)0.5*(d2 + d3 + d4 - d1);
				}
			}
		}
		static void Do2D(T* data,size_t size)
		{
			TransposeQuarters(data,size);
			Do1DRowsAndCols(data,size);
			AddQuadrants(data,size);
			TransposeQuarters(data,size);
		}
	};
	*/
}//namespace IntTransf
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
} //namespace NumAlg
} //namespace Vtl
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
#endif //_INTEGRAL_TRANSFORMS_H
