// file: #include "num_optimization.h"
#ifndef _NUM_OPTIMIZATION_H
#define _NUM_OPTIMIZATION_H
//////////////////////////////////////////////////////////////////////////////////////////////////////
#include "numalg.h"
#include "num_linalg.h"
//////////////////////////////////////////////////////////////////////////////////////////////////////
namespace Vtl {namespace NumAlg { namespace Opt {
	//////////////////////////////////////////////////////////////////////////////////////////////////////
	//
	///
	///         X  ,  fun  
	///       [A,B].
	///
	///     \param[in] fun  -  ,   fun(X)   X   [A,B].
	///     \param[in] A    -     
	///     \param[in] B    -     
	///     \param[in] tol  -        / tol >= 0. /.
	///
	///     \retval Fmin -  ,  ,   fun  .
	///
	///        \ref Lit_Forsythe80 ,       C++   
	///     .  - :\n
	/// \li      
	/// \li        ,       
	///           . 
	/// \li       .
	///
	///               -
	///         .    
	///     ,    .  fun    -
	///     ,    (    A, 
	///      B),        
	///     1.324...
	///         fun      ,   
	///         eps*abs(X)+tol/3 ,  eps  
	///          .  fun - -
	///          fun  
	///         ,  FMIN -
	///         fun   (A,B)  ,
	///      3*eps*abs(X)+tol.  fun   ,   FMIN
	///            , -
	///     ,    .
	///        -    
	///      60 -  LOCALMIN,   Richard Brent , 
	///     Algorithms for minimization without derivatives, Prentice-Hall,inc.
	///     (1973).
	///
	//-///////////////////////////////////////////////////////////////////////////
	template<typename T,class Functor>
	inline T Fmin(Functor fun,T A,T B,T tol=T(0))
	{
		using namespace std; //function std::abs, std::sqrt
		using namespace Vtl::NumAlg; // function sign(,) , analog of FORTRAN SIGN(,)

		//      G     ,    
		const T G=T(0.5*(3.- sqrt(5.)));
		//
		//     eps       -
		//      .
		//
		const T eps = sqrt( RelerrEps<T>() );
		//---------------<    >------
		T v=A+G*(B-A);
		T w=v;
		T x=v;
		T e=0.;
		T fx=fun(x);
		T fv=fx;
		T fw=fx;

		T d,p,q,u;
		//***------------<     >-------
		while(1)
		{
			const T xm=0.5*(A+B);
			const T tol1=eps*abs(x)+tol/3.;
			const T tol2=2.*tol1;
			//===============<    >=========
			if( abs(x-xm) <= ( tol2-0.5*(B-A) ) )
			{//-------------**<    >**-----
				return x;
			};
			//    
			bool logic= ( abs(e) <= tol1 );
			if( ! logic )
			{
				//   
			   T r=(x-w)*(fx-fv);
			   q=(x-v)*(fx-fw);
			   p=(x-v)*q-(x-w)*r;
			   q=2.*(q-r);
			   if( q > 0. ) p=-p;
			   q=abs(q);
			   r=e;
			   e=d;
				//    
			   logic=logic || ( abs(p) >= abs(0.5*q*r) )
			   || ( p <= q*(A-x) ) || ( p >= q*(B-x) );
			}
			if( logic ) 
			{
				//    
				e = ( x >= xm ) ? (A-x) : (B-x) ; 
				d = G * e;
			}
			else
			{
				//    
				d = p / q ;
				u = x + d ;
				// fun       A  B 
			   if( (u-A) < tol2 ) d=sign(tol1,xm-x);
			   if( (B-u) < tol2 ) d=sign(tol1,xm-x);
			}
			// fun       x 
			if( abs(d) >= tol1 )
				u=x+d;
			else
				u=x+sign(tol1,d);
			
			const T fu = fun(u);
			//     A,B,v,w,x .
			if( fu > fx ) 
			{
				if( u < x ) A=u;
				else        B=u;
				if( fu <= fw || w == x ) 
				{
					v=w;
					fv=fw;
					w=u;
					fw=fu;
				}
				else if( fu <= fv || v == x || v == w ) 
				{
					v=u;
					fv=fu;
				}
			}
			else
			{
				if( u >= x ) A=x;
				else         B=x;
				v=w;
				fv=fw;
				w=x;
				fw=fx;
				x=u;
				fx=fu;
			}
		}//end of while(1);
	}
	///
	/// Given an n-dimensional point xold_[1..n], the value of the function and gradient there, 
	/// fold
	/// and g_[1..n], 
	/// and a direction p_[1..n], 
	/// finds a new point x_[1..n] along the direction p_ from
	/// xold_ 
	/// where the function 
	/// func has decreased "suffciently." 
	/// The new function value is returned in f. 
	/// stpmax is an input quantity that limits the length of the steps so that you do not try to
	/// evaluate the function in regions where it is undefined or subject to overflow. 
	/// p_ is usually the Newton direction. 
	/// The output quantity check is false (0) on a normal exit. 
	/// It is true (1) when x_ is too close to xold_. 
	/// In a minimization algorithm, this usually signals convergence and can be ignored. 
	/// However, in a zero-finding algorithm the calling program should check whether the
	/// convergence is spurious. Some "difficult" problems may require double precision in this routine.
	///
	/// Numerical recipes in C ver 2.0, Page 385 
	/// Chapter 9. Root Finding and Nonlinear Sets of Equations
	///
	template<typename T,int n,class Functor>
	inline void nrc_lnsrch(boost::numeric::ublas::c_vector<T,n>& xold_
			   ,T fold
			   ,boost::numeric::ublas::c_vector<T,n>& g_
			   ,boost::numeric::ublas::c_vector<T,n>& p_
			   ,boost::numeric::ublas::c_vector<T,n>& x_
			   ,T& f
			   ,T stpmax
			   ,int& check
			   ,Functor func)
//	#include <math.h>
//	#include "nrutil.h"
//	#define alf 1.0e-4  // Ensures sufficient decrease in function value.
//	#define tolx 1.0e-7 // Convergence criterion on delta x_.
//	void lnsrch(int n, float xold_[], float fold, float g_[], float p_[], float x_[]
//		  , float *f, float stpmax, int *check
//        , float (*func)(float [])
//		)
	{
		using namespace boost::numeric::ublas;
		//::ublas::blas_1::dot()
		using namespace std; //max ( FMAX), abs
		const T alf=1.0e-4f;  // Ensures sufficient decrease in function value.
		const T tolx=1.0e-7f; // Convergence criterion on delta x_.

		T a,alam2,b,disc,f2,rhs1,rhs2,tmplam;

		check=0;
		////// float sum=0; 
		////// for(int i=1; i<=n; i++ ) 
		//////  	sum += p_[i]*p_[i];
		////// sum = sqrt(sum);
		T sum = sqrt( blas_1::dot(p_,p_) );
		
		if(sum > stpmax)
		{
			//////for(int i=1;i<=n;i++) 
			//////	p_[i] *= stpmax/sum; // Scale if attempted step is too big.
			//p_ *= stpmax/sum; - instead next string
			blas_1::scal(p_ , stpmax/sum);
		}
		
		//float slope=0;
		//for(int i=1 ; i<=n ; i++ )
		//	slope += g_[i]*p_[i];
		T  slope=blas_1::dot(g_,p_);
		if( slope >= T(0) ) 
			throw "Roundoff problem in nrc_lnsrch.";

		//float test=0; //Compute Lambda_min.
		//for(int i=1;i<=n;i++) 
		//{
		//	float temp=abs(p_[i]) / max(abs(xold_[i]),float(1.0));
		//	if(temp > test) 
		//		test=temp;
		//}
		T test=0; //Compute Lambda_min.
		{
			for( int i=0 ; i < n ; ++i ) 
			{
				T temp = abs(p_[i]) / std::max( abs(xold_[i]) ,T(1.0) );
				if(temp > test) 
					test=temp;
			}
		}

		//float alamin=tolx/test;
		//float alam=1.0; // Always try full Newton step first.
		T alamin=tolx/test;
		T alam=T(1); // Always try full Newton step first.

		//////int i;
		for(;;) 
		{ //Start of iteration loop.
			//////for( i=1; i<=n; i++) 
			//////	x_[i]=xold_[i]+alam*p_[i];
			blas_1::copy(x_,xold_);
			blas_1::axpy(x_,alam,p_);

			f= func(x_);
			if(alam < alamin) 
			{ // Convergence on delta_x. For zero finding, the calling program should verify the convergence.
				//////for(i=1;i<=n;i++) 
				//////	x_[i]=xold_[i];
				blas_1::copy(x_,xold_);
				
				check=1;
				return;
			} 
			else if(f <= fold+alf*alam*slope) 
				return; // Sufficient function decrease.
			else 
			{ //Backtrack.
				if( alam == T(1) )
					tmplam = -slope/( T(2)*(f-fold-slope)); //First time.
				else 
				{ //Subsequent backtracks.
					rhs1 = f - fold - alam  * slope;
					rhs2 = f2 - fold - alam2 * slope;
					a = ( rhs1/(alam*alam) - rhs2/(alam2*alam2) ) / (alam-alam2);
					b = ( -alam2*rhs1/(alam*alam) + alam*rhs2/(alam2*alam2) ) / (alam-alam2);
					if( a == T(0) ) 
						tmplam = -slope/(T(2)*b);
					else 
					{
						disc = b*b-T(3)*a*slope;
						if(disc < T(0) ) 
							tmplam = T(0.5)*alam;
						else if(b <= T(0) ) 
							tmplam = (-b+sqrt(disc))/(T(3)*a);
						else 
							tmplam = -slope/(b+sqrt(disc));
					}
					if( tmplam > T(0.5)*alam )
					tmplam=T(0.5)*alam; // Lambda <= 0.5*Lambda_1
				}
			}
			alam2=alam;
			f2 = f;
			alam = std::max( tmplam, T(0.1)*alam ); // Lambda >= 0.1*Lambda_1
		} //Try again.
	}
	//////#include <math.h>
	//////#include "nrutil.h"
	//////#define itmax 200 Maximum allowed number of iterations.
	//////#define eps 3.0e-8 Machine precision.
	//////#define tolx (4*eps) Convergence criterion on x_ values.
	//////#define stpmx 100.0 Scaled maximum step length allowed in line searches.
	//////#define FREEALL free_vector(xi_,1,n);free_vector(pnew_,1,n); \
	//////		free_matrix(hessin,1,n,1,n);free_vector(hdg_,1,n);free_vector(g_,1,n); \
	//////		free_vector(dg_,1,n);
	////// template<int n>
	////// float dfpmin(float p_[], float gtol, int *iter
	//////  	   , float(*func)(float [])
	//////         , void (*dfunc)(float [], float []))
	//
	///
	/// Given a starting point 
	/// \param p_ - that is a vector of length n.
	/// the Broyden-Fletcher-Goldfarb-Shanno variant of Davidon-Fletcher-Powell minimization is performed 
	/// on a function 
	/// \param gtol  - The convergence requirement on zeroing the gradient is input as gtol. 
	/// \param func  - functor which specify a function to minimize and a gradient of this function/
	/// Returned quantities are:
	/// \param p_      - the location of the minimum,
	/// \param iter    - the number of iterations that were performed, 
	/// \return        - the minimum value of the function. 
	/// The routine nrc_lnsrch is called to perform approximate line minimizations.
	///
	/// From Numerical Recipes 2.0 in C
	///	10.7 Variable Metric Methods in Multidimensions, page 428
	///
	/// 
	/// Functor   :
	/// \code
	/// template<typename T,int n>
	/// struct Func
	/// {
	///    typedef boost::numeric::ublas::c_vector<T,n> Vector;
	///     // returns function Func value at n-dimentional point v
	///     T operator(const Vector& v )() 
	///     {
	///        //some code.
	///     }
	///     // if You whant, You can add this operator (but it is not required!)
	///     T operator(const Vector& v,Loki::Int2type<0> )()
	///     {
	///         retunr (*this)(v);
	///     }
	///     // returns function Func GRADIENT value (vector with 
	///     // componet i=0..(n-1) wich is 1-st derivative in v[i] ) at n-dimentional point v
	///     Vector operator(const Vector& v,Loki::Int2type<1> )() 
	///     {
	///        //Implementation MUST BE such as (for efficiency reasons):
	///			Vector grad_vector;
	///			grad(grad_vector,v); 
	///			return grad_vector;
	///     }
	/// private:
	///     // returns function Func GRADIENT value (vector with 
	///     // componet i=0..(n-1) wich is 1-st derivative in v[i] ) at n-dimentional point v
	///     void grad(Vector& grad_vector,const Vector& v) 
	///     {
	///        //some code.
	///     }
	/// };
	/// \endcode
	///
	template<typename T,int n,class Functor>
	inline T nrc_dfpmin(boost::numeric::ublas::c_vector<T,n>& p_
				,T gtol, int& iter 
				,Functor func)
	{
		using namespace std; //max ( FMAX), abs ( fabs) , sqrt
		using namespace boost::numeric::ublas;
		typedef boost::numeric::ublas::c_vector<T,n> Vector;
		typedef boost::numeric::ublas::c_matrix<T,n,n> Matrix;
		
		enum {itmax = 200};// Maximum allowed number of iterations.
		const T eps=Vtl::NumAlg::RelerrEps<T>();//3.0e-8f; // Machine precision.
		const T tolx=4.0f*eps; // Convergence criterion on x_ values.
		const T stpmx=100.0f; //Scaled maximum step length allowed in line searches.
		//void lnsrch(int n, float xold_[], float fold, float g_[], float p_[], float x_[],
		//			float *f, float stpmax, int *check, float (*func)(float []));
		T fret;

		//////int i,j;
		int check,its;
		T den,fac,fad,fae,fp,stpmax;
		T sumdg,sumxi,temp,test;

		//////float *dg_,*g_,*hdg_,*pnew_,*xi_,**hessin;
		//////dg_=vector(1,n);
		//////g_=vector(1,n);
		//////hdg_=vector(1,n);
		//////pnew_=vector(1,n);
		//////xi_=vector(1,n);
		//////hessin=matrix(1,n,1,n);
		Vector dg_;
		Vector g_;
		Vector hdg_;
		Vector pnew_;
		Vector xi_;
		Matrix hessin;

		//////fp=(*func)(p_); // Calculate starting function value 
		//////(*dfunc)(p_,g_); // and gradient,
		fp=func(p_); // Calculate starting function value 
		g_=func(p_,Loki::Int2Type<1>()); // and gradient,

		//////float sum=0.0f;
		//////for(i=1;i<=n;i++) 
		//////{ //and initialize the inverse Hessian to the unit matrix.
		//////	for (j=1;j<=n;j++) hessin[i][j]=0.0; 
		//////		hessin[i][i]=1.0;
		//////	xi_[i] = -g_[i]; //Initial line direction.
		//////	sum += p_[i]*p_[i];
		//////}
		//and initialize the inverse Hessian to the unit matrix.
		hessin = identity_matrix<T>(n,n);
		xi_ = -g_; //Initial line direction.
		T sum=blas_1::dot(p_,p_);

		stpmax = stpmx * std::max(sqrt(sum),T(n));
		for(its=1;its<=itmax;its++) 
		{ //Main loop over the iterations.
			iter=its;
			nrc_lnsrch(p_,fp,g_,xi_,pnew_,fret,stpmax,check,func);
			//The new function evaluation occurs in nrc_lnsrch; save the function value in fp for the
			//next line search. It is usually safe to ignore the value of check.
			fp = fret;
			//////for(i=1;i<=n;i++) 
			//////{
			//////	xi_[i]=pnew_[i]-p_[i];  // Update the line direction,
			//////	p_[i]=pnew_[i];        // and the current point.
			//////}
			xi_=pnew_-p_;  // Update the line direction,
			p_=pnew_;        // and the current point.

			test=T(0); //Test for convergence on Delta x_.
			{
				for(int i=0 ; i < n ; ++i ) 
				{
					temp = abs(xi_[i]) / std::max( abs(p_[i]) , T(1) );
					if( temp > test ) test=temp;
				}
			}
			if( test < tolx ) 
			{
				//FREEALL
				return fret;
			}

			//////for (i=1;i<=n;i++) dg_[i]=g_[i]; // Save the old gradient,
			blas_1::copy(dg_,g_);

			//////(*dfunc)(p_,g_); //and get the new gradient.
			g_=func(p_,Loki::Int2Type<1>());

			test=T(0); //Test for convergence on zero gradient.
			den=std::max(fret,T(1));
			for(int i=0 ; i < n ; ++i ) 
			{
				T temp = abs(g_[i]) * std::max(abs(p_[i]),T(1)) / den;
				if (temp > test) test=temp;
			}
			if( test < gtol ) 
			{
				//FREEALL
				return fret;
			}
			//////for (i=1;i<=n;i++) dg_[i]=g_[i]-dg_[i]; // Compute difference of gradients,
			dg_ = g_ - dg_; // Compute difference of gradients,

			//////for (i=1;i<=n;i++) 
			//////{ // and difference times current matrix.
			//////	hdg_[i]=0.0;
			//////	for (j=1;j<=n;j++) 
			//////		hdg_[i] += hessin[i][j]*dg_[j];
			//////}
			// and difference times current matrix.
			axpy_prod(hessin,dg_,hdg_,true); // hdg_  = prod(hessin,dg_) ; //param true means that v.clear before axpy_prod, so axpy_prod(A,x,v) v+=Ax becomes v=Ax
			
			//////fac=fae=sumdg=sumxi=0.0; //Calculate dot products for the denominators.
			//////for(i=1;i<=n;i++) 
			//////{ 
			//////	fac += dg_[i]*xi_[i];
			//////	fae += dg_[i]*hdg_[i];
			//////	sumdg += Vtl::sqr(dg_[i]);
			//////	sumxi += Vtl::sqr(xi_[i]);
			//////}
			//Calculate dot products for the denominators.
			fac = blas_1::dot(dg_,xi_);
			fae = blas_1::dot(dg_,hdg_);
			//sumdg = blas_1::dot(dg_,dg_);
			//sumxi = blas_1::dot(xi_,xi_);
			sumdg = blas_1::nrm2(dg_);
			sumxi = blas_1::nrm2(xi_);

			if( fac > sqrt(eps*sumdg*sumxi) ) 
			{ //Skip update if fac not sufficiently positive.
				fac=T(1)/fac; 
				fad=T(1)/fae;
				// The vector that makes BFGS different from DFP:
				//////for(i=1;i<=n;i++) 
				//////	dg_[i]=fac*xi_[i]-fad*hdg_[i];
				dg_ = fac*xi_ - fad*hdg_; //????

				//////for(i=1;i<=n;i++) 
				//////{ //The BFGS updating formula:
				//////	for(j=i;j<=n;j++) 
				//////	{
				//////		hessin[i][j] += fac * xi_[i]*xi_[j] - fad * hdg_[i]*hdg_[j] + fae * dg_[i]*dg_[j];
				//////		hessin[j][i]=hessin[i][j];
				//////	}
				//////}
				//The BFGS updating formula:
				// M blas_2::sr(M &  m,const T& t,const V& v)//symmetric rank 1 update: m = m + t * (v * vT) 
				blas_2::sr(hessin, fac,xi_ );
				blas_2::sr(hessin,-fad,hdg_);
				blas_2::sr(hessin, fae,dg_ );
			}
			//////for (i=1;i<=n;i++) 
			//////{ //Now calculate the next direction to go,
			//////	xi_[i]=0.0;
			//////	for (j=1;j<=n;j++) 
			//////		xi_[i] -= hessin[i][j]*g_[j];
			//////}
			//Now calculate the next direction to go,
			axpy_prod(hessin,g_,xi_,true); // xi_ = prod(hessin,g_) ; //param true means that v.clear before axpy_prod, so axpy_prod(A,x,v) v+=Ax becomes v=Ax
			xi_ = - xi_;

		}// and go back for another iteration.
		throw "too many iterations in dfpmin";
		//FREEALL
	}
//////////////////////////////////////////////////////////////////////////////////////////////////////
}}} // namespace Vtl {namespace NumAlg { namespace Opt {
//////////////////////////////////////////////////////////////////////////////////////////////////////
#endif //_NUM_OPTIMIZATION_H
